"""network_importer specific diff class based on diffsync.

(c) 2020 Network To Code

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
  http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from collections import defaultdict
from diffsync.diff import Diff


class NetworkImporterDiff(Diff):
    """network_importer specific diff class based on diffsync."""

    @classmethod
    def order_children_interface(cls, children):
        """Return the interface children ordered order."""
        intfs_lags = defaultdict(list)
        intfs_regs = defaultdict(list)
        intfs_lag_members = defaultdict(list)

        for child_name, child in children.items():
            action = child.action

            if action is None:
                action = "update"

            if action == "delete":
                if "is_lag" in child.dest_attrs and child.dest_attrs["is_lag"]:
                    intfs_lags[action].append(child_name)
                elif "is_lag_member" in child.dest_attrs and child.dest_attrs["is_lag_member"]:
                    intfs_lag_members[action].append(child_name)
                else:
                    intfs_regs[action].append(child_name)

            elif action in ["update", "create"]:

                if "is_lag" in child.source_attrs and child.source_attrs["is_lag"]:
                    intfs_lags[action].append(child_name)
                elif "is_lag_member" in child.source_attrs and child.source_attrs["is_lag_member"]:
                    intfs_lag_members[action].append(child_name)
                else:
                    intfs_regs[action].append(child_name)

            else:
                raise Exception("invalid DiffElement")

        sorted_intfs = intfs_regs["create"]
        sorted_intfs += intfs_regs["update"]
        sorted_intfs += intfs_lags["create"]
        sorted_intfs += intfs_lags["update"]
        sorted_intfs += intfs_lag_members["create"]
        sorted_intfs += intfs_lag_members["update"]
        sorted_intfs += intfs_regs["delete"]
        sorted_intfs += intfs_lags["delete"]
        sorted_intfs += intfs_lag_members["delete"]

        for intf in sorted_intfs:
            yield children[intf]



"""Collection of Nornir tasks for the network importer.

(c) 2020 Network To Code

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
  http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import os
import socket
from typing import Optional, List

import yaml
from nornir.core.task import Result, Task

import network_importer.config as config

LOGGER = logging.getLogger("network-importer")  # pylint: disable=C0103


def device_save_hostvars(task: Task) -> Result:
    """Save the device hostvars into a yaml file.

    Args:
      task (Task): Nornir Task

    Returns:
      Result: Nornir Result
    """
    if not task.host.data["obj"].hostvars:
        return Result(host=task.host)

    # Save device hostvars in file
    if not os.path.exists(f"{config.SETTINGS.main.hostvars_directory}/{task.host.name}"):
        os.makedirs(f"{config.SETTINGS.main.hostvars_directory}/{task.host.name}")
        LOGGER.debug("Directory %s/%s was missing, created it", config.SETTINGS.main.hostvars_directory, task.host.name)

    with open(
        f"{config.SETTINGS.main.hostvars_directory}/{task.host.name}/network_importer.yaml",
        "w",
    ) as out_file:
        out_file.write(yaml.dump(task.host.data["obj"].hostvars, default_flow_style=False))
        LOGGER.debug(
            "%s - Host variables saved in %s/%s/network_importer.yaml",
            task.host.name,
            config.SETTINGS.main.hostvars_directory,
            task.host.name,
        )

    return Result(host=task.host)

    # -------------------------------------------------------------------
    # Old code that need used previously to generate the hostvars from a jinja templates
    # -------------------------------------------------------------------
    # module_path = os.path.dirname(network_importer.__file__)
    # template_dir = f"{module_path}/templates/"

    # dev_facts = task.host.data["obj"].bf.extract_facts(nodes=task.host.name)["nodes"][task.host.name]
    # del dev_facts["Interfaces"]

    # # Load Jinja2 template
    # # env = Environment(
    # #     loader=FileSystemLoader(template_dir), trim_blocks=True, lstrip_blocks=True
    # # )
    # # env.filters["to_yaml_list"] = jinja_filter_toyaml_list
    # # env.filters["to_yaml_dict"] = jinja_filter_toyaml_dict
    # # template = env.get_template("hostvars.j2")
    # # hostvars_str = template.render(dev_facts)


def tcp_ping(task: Task, ports: List[int], timeout: int = 2, host: Optional[str] = None) -> Result:
    """Tests connection to a tcp port and tries to establish a three way handshake.

    Arguments:
        ports (list of int): tcp ports to ping
        timeout (int, optional): defaults to 2
        host (string, optional): defaults to ``hostname``

    Returns:
        Result object with the following attributes set:
          * result (``dict``): Contains port numbers as keys with True/False as values

    Code copied from https://github.com/nornir-automation/nornir/blob/v2.5.0/nornir/plugins/tasks/networking/tcp_ping.py
    Need to open a PR to https://github.com/nornir-automation/nornir_utils
    """
    if isinstance(ports, int):
        ports = [ports]

    if isinstance(ports, list):
        if not all(isinstance(port, int) for port in ports):
            raise ValueError("Invalid value for 'ports'")

    else:
        raise ValueError("Invalid value for 'ports'")

    host = host or task.host.hostname

    result = {}
    for port in ports:
        skt = socket.socket()
        skt.settimeout(timeout)
        try:
            status = skt.connect_ex((host, port))
            if status == 0:  # pylint: disable=simplifiable-if-statement
                connection = True
            else:
                connection = False
        except (socket.gaierror, socket.timeout, socket.error):
            connection = False
        finally:
            skt.close()
        result[port] = connection

    return Result(host=task.host, result=result)


def check_if_reachable(task: Task) -> Result:
    """Check if a device is reachable by doing a TCP ping it on port 22.

    Will change the status of the variable `host.is_reachable` based on the results

    Args:
      task: Nornir Task

    Returns:
       Result: Nornir Result
    """
    port_to_check = 22
    try:
        results = task.run(task=tcp_ping, ports=[port_to_check])
    except:  # noqa: E722 # pylint: disable=bare-except
        LOGGER.debug(
            "An exception occured while running the reachability test (tcp_ping)",
            exc_info=True,
        )
        return Result(host=task.host, failed=True)

    is_reachable = results[0].result[port_to_check]

    if not is_reachable:
        LOGGER.debug("%s | device is not reachable on port %s", task.host.name, port_to_check)
        task.host.is_reachable = False
        task.host.not_reachable_reason = f"device not reachable on port {port_to_check}"
        task.host.status = "fail-ip"

    return Result(host=task.host, result=is_reachable)


def warning_not_reachable(task: Task) -> Result:
    """Generate warning logs for each unreachable device."""
    if task.host.is_reachable:
        return

    reason = task.host.not_reachable_reason or "reason not defined"
    LOGGER.warning("%s device is not reachable, %s", task.host.name, reason)



"""main cli for the network importer.

(c) 2020 Network To Code

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
  http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
# pylint: disable=C0103,W0102,C0415,R1724,W0611,R1710,R1710,E1101,W0613,,C0413,R0904,E1120

import logging
import sys
import pdb

import click
import urllib3
from diffsync.logging import enable_console_logging

from rich.console import Console
from rich.table import Table

import network_importer.config as config
from network_importer.main import NetworkImporter
from network_importer.inventory import reachable_devs
from network_importer.tasks import check_if_reachable

import network_importer.performance as perf

urllib3.disable_warnings()

__author__ = "Damien Garros <damien.garros@networktocode.com>"

LOGGER = logging.getLogger("network-importer")

# pylint: disable=too-many-arguments


@click.group()
def main():
    """Main CLI command for the network_importer."""


def init(config_file):
    """Init Network-Importer."""
    config.load_and_exit(config_file_name=config_file)
    perf.init()

    # ------------------------------------------------------------
    # Setup Logging
    # ------------------------------------------------------------
    logging.getLogger("pybatfish").setLevel(logging.ERROR)
    logging.getLogger("netaddr").setLevel(logging.ERROR)

    if config.SETTINGS.logs.level != "debug":
        logging.getLogger("paramiko.transport").setLevel(logging.CRITICAL)
        logging.getLogger("nornir.core.task").setLevel(logging.CRITICAL)

    logging.basicConfig(stream=sys.stdout, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s")

    if config.SETTINGS.logs.level == "debug":
        LOGGER.setLevel(logging.DEBUG)
    elif config.SETTINGS.logs.level == "warning":
        LOGGER.setLevel(logging.WARNING)
    else:
        LOGGER.setLevel(logging.INFO)

    # Disable logging in console for DiffSync
    enable_console_logging(verbosity=0)

    ni = NetworkImporter()
    return ni


@click.option(
    "--config",
    "config_file",
    default="network_importer.toml",
    help="Network Importer Configuration file (TOML format)",
    type=str,
    show_default=True,
)
@click.option(
    "--limit",
    default=False,
    help="limit the execution on a specific device or group of devices --limit=device1 or --limit='site=sitea' ",
    type=str,
)
@click.option("--update-configs", is_flag=True, help="Pull the latest configs from the devices")
@click.option(
    "--debug", is_flag=True, help="Keep the script in interactive mode once finished for troubleshooting", hidden=True
)
@main.command()
def apply(config_file, limit, debug, update_configs):
    """Save changes in Backend."""
    ni = init(config_file)

    if update_configs:
        ni.build_inventory(limit=limit)
        ni.update_configurations()

    ni.init(limit=limit)
    ni.sync()

    perf.TIME_TRACKER.set_nbr_devices(len(ni.nornir.inventory.hosts.keys()))
    if config.SETTINGS.logs.performance_log:
        perf.TIME_TRACKER.print_all()

    LOGGER.info("Execution finished, processed %s device(s)", perf.TIME_TRACKER.nbr_devices)
    if debug:
        pdb.set_trace()

    ni.sync()


@click.option(
    "--config",
    "config_file",
    default="network_importer.toml",
    help="Network Importer Configuration file (TOML format)",
    type=str,
    show_default=True,
)
@click.option(
    "--limit",
    default=False,
    help="limit the execution on a specific device or group of devices --limit=device1 or --limit='site=sitea' ",
    type=str,
)
@click.option("--update-configs", is_flag=True, help="Pull the latest configs from the devices")
@click.option(
    "--debug", is_flag=True, help="Keep the script in interactive mode once finished for troubleshooting", hidden=True
)
@main.command()
def check(config_file, limit, debug, update_configs):
    """Display what are the differences but do not save them."""
    ni = init(config_file)

    if update_configs:
        ni.build_inventory(limit=limit)
        ni.update_configurations()

    ni.init(limit=limit)
    diff = ni.diff()
    print(diff.str())

    perf.TIME_TRACKER.set_nbr_devices(len(ni.nornir.inventory.hosts.keys()))
    if config.SETTINGS.logs.performance_log:
        perf.TIME_TRACKER.print_all()

    LOGGER.info("Execution finished, processed %s device(s)", perf.TIME_TRACKER.nbr_devices)
    if debug:
        pdb.set_trace()


@click.option(
    "--config",
    "config_file",
    default="network_importer.toml",
    help="Network Importer Configuration file (TOML format)",
    type=str,
    show_default=True,
)
@click.option(
    "--limit",
    default=False,
    help="limit the execution on a specific device or group of devices --limit=device1 or --limit='site=sitea' ",
    type=str,
)
@click.option("--check-connectivity", is_flag=True, help="Check if the devices are reachable on port 22")
@click.option("--update-configs", is_flag=True, help="Pull the latest configs from the devices")
@click.option(
    "--debug", is_flag=True, help="Keep the script in interactive mode once finished for troubleshooting", hidden=True
)
@main.command()
def inventory(config_file, limit, debug, check_connectivity, update_configs):
    """Display inventory."""
    ni = init(config_file)
    ni.build_inventory(limit=limit)

    if check_connectivity:
        ni.nornir.filter(filter_func=reachable_devs).run(task=check_if_reachable, on_failed=True)

    if update_configs:
        ni.update_configurations()

    if limit:
        table = Table(title=f"Device Inventory (limit:{limit})")
    else:
        table = Table(title="Device Inventory (all)")

    table.add_column("Device", style="cyan", no_wrap=True)
    table.add_column("Platform", style="magenta")
    table.add_column("Driver")
    table.add_column("Reachable")
    table.add_column("Reason")

    for hostname, host in ni.nornir.inventory.hosts.items():
        if host.is_reachable:
            is_reachable = "[green]True"
            reason = None
        else:
            is_reachable = "[red]False"
            reason = f"[red]{host.not_reachable_reason}"

        driver = config.SETTINGS.drivers.mapping.get(host.platform, config.SETTINGS.drivers.mapping.get("default"))
        table.add_row(hostname, host.platform, driver, is_reachable, reason)

    console = Console()
    console.print(table)

    if debug:
        pdb.set_trace()


if __name__ == "__main__":
    main()



"""default network_importer driver for cisco.

(c) 2020 Network To Code

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
  http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging

from network_importer.drivers.default import NetworkImporterDriver as DefaultNetworkImporterDriver

LOGGER = logging.getLogger("network-importer")


class NetworkImporterDriver(DefaultNetworkImporterDriver):
    """Collection of Nornir Tasks specific to Juniper Junos devices."""



"""Base Inventory and Host class for Network Importer."""

from typing import Dict, List, Optional

from nornir.core.inventory import Host, Group, ConnectionOptions

# pylint: disable=too-many-arguments,too-many-instance-attributes


class NetworkImporterHost(Host):
    """Network Importer Host Class."""

    site_name: Optional[str]
    """Name of the site this device belong to."""

    is_reacheable: Optional[bool]
    """Global Flag to indicate if we are able to connect to a device"""

    status: Optional[str] = "ok"
    """ Valid Statuses
        ok: device is reachable
        fail-ip: Primary IP address not reachable
        fail-access: Unable to access the device management.
                     The IP is reachable, but SSH or API is not enabled or responding.
        fail-login: Unable to login authenticate with device
        fail-other:  Other general processing error (also catches traps/bug)
    """

    has_config: Optional[bool] = False
    """ Indicate if the configuration is present and has been properly imported in Batfish."""

    not_reachable_reason: Optional[str]


class NetworkImporterInventory:
    """Base inventory class for the Network Importer."""

    def __init__(
        self,
        username: Optional[str] = None,
        password: Optional[str] = None,
        enable: Optional[bool] = None,
        supported_platforms: Optional[List[str]] = None,
        netmiko_extras: Optional[Dict] = None,
        napalm_extras: Optional[Dict] = None,
        limit: Optional[str] = None,
        settings: Optional[Dict] = None,
    ):
        """Initialize class and store all top level arguments locally."""
        self.username = username
        self.password = password
        self.enable = enable
        self.supported_platforms = supported_platforms
        self.limit = limit
        self.netmiko_extras = netmiko_extras
        self.napalm_extras = napalm_extras

        self.settings = settings

        # Define Global Group with Netmiko and Napalm Credentials if provided
        self.global_group = Group(
            name="global", connection_options={"netmiko": ConnectionOptions(), "napalm": ConnectionOptions()}
        )

        if self.netmiko_extras:
            self.global_group.connection_options["netmiko"].extras = self.netmiko_extras

        if self.napalm_extras:
            self.global_group.connection_options["napalm"].extras = self.napalm_extras

        # Pull the login and password from the NI config object if available
        if self.username:
            self.global_group.username = self.username

        if self.password:
            self.global_group.password = self.password
            if self.enable:
                if not self.global_group.connection_options["netmiko"].extras:
                    self.global_group.connection_options["netmiko"].extras = dict()
                elif "secret" not in self.global_group.connection_options["netmiko"].extras:
                    self.global_group.connection_options["netmiko"].extras["secret"] = self.password

                if not self.global_group.connection_options["napalm"].extras:
                    self.global_group.connection_options["napalm"].extras = {"optional_args": {"secret": self.password}}
                elif isinstance(dict, self.global_group.connection_options["napalm"].extras):
                    if "optional_args" not in self.global_group.connection_options["napalm"].extras:
                        self.global_group.connection_options["napalm"].extras["optional_args"] = {
                            "secret": self.password
                        }
                    elif isinstance(dict, self.global_group.connection_options["napalm"].extras["optional_args"]):
                        self.global_group.connection_options["napalm"].extras["optional_args"]["secret"] = self.password


# -----------------------------------------------------------------
# Inventory Filter functions
# -----------------------------------------------------------------
def valid_devs(host):
    """Inventory Filter for Nornir for all valid devices.

    Return True or False if a device is valid

    Args:
      host(Host): Nornir Host

    Returns:
        bool: True if the device has a config, False otherwise.
    """
    if host.has_config:
        return True

    return False


def non_valid_devs(host):
    """Inventory Filter for Nornir for all non-valid devices.

    Return True or False if a device is not valid

    Args:
      host(Host): Nornir Host

    Returns:
        bool: True if the device do not have a config, False otherwise.
    """
    if host.has_config:
        return False

    return True


def reachable_devs(host):
    """Inventory Filter for Nornir for all reachable devices.

    Return True if the device is reachable.

    Args:
      host(Host): Nornir Host

    Returns:
        bool: True if the device is reachable, False otherwise.
    """
    if host.is_reachable:
        return True

    return False


def non_reachable_devs(host):
    """Inventory Filter for Nornir for all non reachable devices.

    Return True if the device is not reachable.

    Args:
      host(Host): Nornir Host

    Returns:
        bool: True if the device is not reachable, False otherwise.
    """
    if host.is_reachable:
        return False

    return True


def valid_and_reachable_devs(host):
    """Inventory Filter for Nornir for all valid and reachable devices.

    Return True if the device is reachable and has a config.

    Args:
      host(Host): Nornir Host

    Returns:
        bool: True if the device is reachable and has a config, False otherwise.
    """
    if host.is_reachable and host.has_config:
        return True

    return False



"""network_importer driver for arista_eos.

(c) 2020 Network To Code

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
  http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import logging

from nornir.core.task import Result, Task

from network_importer.drivers.default import NetworkImporterDriver as DefaultNetworkImporterDriver
from network_importer.processors.get_vlans import Vlan, Vlans

LOGGER = logging.getLogger("network-importer")


class NetworkImporterDriver(DefaultNetworkImporterDriver):
    """Collection of Nornir Tasks specific to Arista EOS devices."""

    @staticmethod
    def get_vlans(task: Task) -> Result:
        """Get a list of vlans from the device.

        Args:
            task (Task): Nornir Task

        Returns:
            Result: Nornir Result object with a dict as a result containing the vlans
            The format of the result but must be similar to Vlans defined in network_importer.processors.get_vlans
        """
        results = Vlans()

        nr_device = task.host.get_connection("napalm", task.nornir.config)
        eos_device = nr_device.device
        results = eos_device.run_commands(["show vlan"])

        if not isinstance(results[0], dict) or "vlans" not in results[0]:
            LOGGER.warning("%s | No vlans information returned", task.host.name)
            return Result(host=task.host, result=False)

        for vid, data in results[0]["vlans"].items():
            results.vlans.append(Vlan(name=data["name"], id=vid))

        return Result(host=task.host, result=results)



"""main class for the network importer."""
import logging
import os
import sys
import warnings
import importlib

from pydantic import ValidationError

import network_importer.config as config
from network_importer.exceptions import AdapterLoadFatalError
from network_importer.utils import patch_http_connection_pool
from network_importer.processors.get_config import GetConfig
from network_importer.drivers import dispatcher
from network_importer.diff import NetworkImporterDiff
from network_importer.tasks import check_if_reachable, warning_not_reachable
from network_importer.performance import timeit
from network_importer.inventory import reachable_devs

warnings.filterwarnings("ignore", category=DeprecationWarning)

with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    from nornir import InitNornir
    from nornir.core.plugins.inventory import InventoryPluginRegister

__author__ = "Damien Garros <damien.garros@networktocode.com>"


LOGGER = logging.getLogger("network-importer")


class NetworkImporter:
    """Main NetworkImporter object to track all state related to the network importer."""

    def __init__(self, check_mode=False, nornir=None):
        """Initialize the NetworkImporter class."""
        self.nornir = nornir
        self.check_mode = check_mode
        self.network = None
        self.sot = None

    @timeit
    def build_inventory(self, limit=None):
        """Build the inventory for the Network Importer in Nornir format."""
        # pylint: disable=import-outside-toplevel
        # Load build-in Inventories as needed
        if config.SETTINGS.inventory.inventory_class == "NetBoxAPIInventory":
            from network_importer.adapters.netbox_api.inventory import NetBoxAPIInventory

            InventoryPluginRegister.register("NetBoxAPIInventory", NetBoxAPIInventory)
        elif config.SETTINGS.inventory.inventory_class == "NautobotAPIInventory":
            from network_importer.adapters.nautobot_api.inventory import NautobotAPIInventory

            InventoryPluginRegister.register("NautobotAPIInventory", NautobotAPIInventory)

        self.nornir = InitNornir(
            runner={"plugin": "threaded", "options": {"num_workers": config.SETTINGS.main.nbr_workers}},
            logging={"enabled": False},
            inventory={
                "plugin": config.SETTINGS.inventory.inventory_class,
                "options": {
                    "username": config.SETTINGS.network.login,
                    "password": config.SETTINGS.network.password,
                    "enable": config.SETTINGS.network.enable,
                    "supported_platforms": config.SETTINGS.inventory.supported_platforms,
                    "netmiko_extras": config.SETTINGS.network.netmiko_extras,
                    "napalm_extras": config.SETTINGS.network.napalm_extras,
                    "limit": limit,
                    "settings": config.SETTINGS.inventory.settings,
                },
            },
        )

        return True

    @timeit
    def init(self, limit=None):
        """Initialize NetworkImporter Object.

        Args:
          limit (str): (Default value = None)
        """
        patch_http_connection_pool(maxsize=100)

        if not self.nornir:
            self.build_inventory(limit=limit)

        # --------------------------------------------------------
        # Creating required directories on local filesystem
        # --------------------------------------------------------
        if (
            not self.check_mode
            and not os.path.exists(config.SETTINGS.main.hostvars_directory)
            and config.SETTINGS.main.generate_hostvars
        ):
            os.makedirs(config.SETTINGS.main.hostvars_directory)
            LOGGER.debug("Directory %s was missing, created it", config.SETTINGS.main.hostvars_directory)

        # --------------------------------------------------------
        # Initialize Adapters
        # --------------------------------------------------------
        LOGGER.info("Import SOT Model")
        sot_path = config.SETTINGS.adapters.sot_class.split(".")
        sot_settings = config.SETTINGS.adapters.sot_settings
        sot_adapter = getattr(importlib.import_module(".".join(sot_path[0:-1])), sot_path[-1])

        try:
            self.sot = sot_adapter(nornir=self.nornir, settings=sot_settings)
            self.sot.load()
        except ValidationError as exc:
            print(f"Configuration not valid, found {len(exc.errors())} error(s)")
            for error in exc.errors():
                print(f"  {'/'.join(error['loc'])} | {error['msg']} ({error['type']})")
            sys.exit(1)
        except AdapterLoadFatalError as exc:
            LOGGER.error("Unable to load the SOT Adapter : %s", exc)
            sys.exit(1)

        LOGGER.info("Import Network Model")
        network_adapter_path = config.SETTINGS.adapters.network_class.split(".")
        network_adapter_settings = config.SETTINGS.adapters.network_settings
        network_adapter = getattr(
            importlib.import_module(".".join(network_adapter_path[0:-1])), network_adapter_path[-1]
        )
        try:
            self.network = network_adapter(nornir=self.nornir, settings=network_adapter_settings)
            self.network.load()
        except ValidationError as exc:
            print(f"Configuration not valid, found {len(exc.errors())} error(s)")
            for error in exc.errors():
                print(f"  {'/'.join(error['loc'])} | {error['msg']} ({error['type']})")
            sys.exit(1)
        except AdapterLoadFatalError as exc:
            LOGGER.error("Unable to load the SOT Adapter : %s", exc)
            sys.exit(1)

        return True

    def sync(self):
        """Synchronize the SOT adapter and the network adapter."""
        self.sot.sync_from(self.network, diff_class=NetworkImporterDiff)

    def diff(self):
        """Generate a diff of the SOT adapter and the network adapter."""
        return self.sot.diff_from(self.network, diff_class=NetworkImporterDiff)

    @timeit
    def update_configurations(self):
        """Pull the latest configurations from all reachable devices.

        Automatically cleanup the directory after to remove all configurations that have not been updated
        """
        LOGGER.info("Updating configuration from devices .. ")

        # ----------------------------------------------------
        # Do a pre-check to ensure that all devices are reachable
        # ----------------------------------------------------
        self.nornir.filter(filter_func=reachable_devs).run(task=check_if_reachable, on_failed=True)
        self.nornir.filter(filter_func=reachable_devs).run(task=warning_not_reachable, on_failed=True)

        self.nornir.filter(filter_func=reachable_devs).with_processors([GetConfig()]).run(
            task=dispatcher,
            method="get_config",
            on_failed=True,
        )

        return True



"""BaseAdapter for the network importer."""
from diffsync import DiffSync
from diffsync.exceptions import ObjectNotFound
from network_importer.models import Site, Device, Interface, IPAddress, Cable, Vlan, Prefix


class BaseAdapter(DiffSync):
    """Base Adapter for the network importer."""

    site = Site
    device = Device
    interface = Interface
    ip_address = IPAddress
    cable = Cable
    vlan = Vlan
    prefix = Prefix

    settings_class = None
    settings = None

    def __init__(self, nornir, settings):
        """Initialize the base adapter and store the Nornir object locally."""
        super().__init__()
        self.nornir = nornir
        self.settings = self._validate_settings(settings)

    def _validate_settings(self, settings):
        """Load and validate the configuration based on the settings_class."""
        if self.settings_class:
            if settings and isinstance(settings, dict):
                return self.settings_class(**settings)  # pylint: disable=not-callable

            return self.settings_class()  # pylint: disable=not-callable

        return settings

    def load(self):
        """Load the local cache with data from the remove system."""
        raise NotImplementedError

    def get_or_create_vlan(self, vlan, site=None):
        """Check if a vlan already exist before creating it. Returns the existing object if it already exist.

        Args:
            vlan (Vlan): Vlan object
            site (Site, optional): Site Object. Defaults to None.

        Returns:
            (Vlan, bool): return a tuple with the vlan and a bool to indicate of the vlan was created or not
        """
        modelname = vlan.get_type()
        uid = vlan.get_unique_id()

        try:
            return self.get(modelname, uid), False
        except ObjectNotFound:
            pass

        self.add(vlan)
        if site:
            site.add_child(vlan)

        return vlan, True

    def get_or_add(self, obj):
        """Add a new object or retrieve it if it already exists.

        Args:
            obj (DiffSyncModel): DiffSyncModel oject

        Returns:
            DiffSyncModel: DiffSyncObject retrieved from the datastore
            Bool: True if the object was created
        """
        modelname = obj.get_type()
        uid = obj.get_unique_id()

        try:
            return self.get(modelname, uid), False
        except ObjectNotFound:
            pass

        self.add(obj)

        return obj, True



"""TimeTracker class to track the performance of the application.

(c) 2020 Network To Code

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
  http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import logging
import math
from time import strftime, time
import network_importer.config as config

TIME_TRACKER = None
LOGGER = logging.getLogger("network-importer")  # pylint: disable=C0103

# pylint: disable=global-statement


def init():
    """Initialize the global time tracker."""
    global TIME_TRACKER

    if not TIME_TRACKER:
        TIME_TRACKER = TimeTracker()


def print_from_ms(msec) -> str:
    """Result time in human readable format from milliseconds.

    Args:
      ms: time in millisecond

    Returns:
        str: Time in human readable format
    """
    ms_per_sec = 1000
    ms_per_min = ms_per_sec * 60

    minutes = math.floor((msec / ms_per_min))
    seconds = math.floor(((msec - (ms_per_min * minutes)) / ms_per_sec))
    millis = msec - (ms_per_min * minutes) - (ms_per_sec * seconds)

    if minutes == 0 and seconds == 0:
        return f"{millis}ms"
    if minutes == 0:
        return f"{seconds}s {millis}ms"

    return f"{minutes}m {seconds}s {millis}ms"


def timeit(method):
    """Decorator to record the execution time of a function."""
    global TIME_TRACKER  # pylint: disable=global-variable-not-assigned

    def timed(*args, **kw):
        """Decorator to record the execution time of a function and store the result in TIME_TRACKER."""
        timestart = time()
        result = method(*args, **kw)
        timeend = time()

        name = method.__name__.upper()
        exec_time = int((timeend - timestart) * 1000)

        if TIME_TRACKER:
            TIME_TRACKER.times[name] = exec_time

        return result

    return timed


class TimeTracker:
    """TimeTracker object used to keep track of different information around the execution of network importer."""

    def __init__(self):
        """Initialize the TimeTracker object."""
        self.start_time = time()
        self.times = {}
        self.nbr_devices = None

    def set_nbr_devices(self, nbr: int):
        """Define the number of devices."""
        self.nbr_devices = nbr

    def print_all(self):
        """Print all information related to time tracking to file if enabled in the configuration."""
        if not os.path.exists(config.SETTINGS.logs.performance_log_directory):
            os.makedirs(config.SETTINGS.logs.performance_log_directory)
            LOGGER.debug("Directory %s was missing, created it", config.SETTINGS.logs.performance_log_directory)

        perflog_filename = strftime("%Y-%m-%d_%H-%M-%S.log")
        perflog_file_path = config.SETTINGS.logs.performance_log_directory + "/" + perflog_filename

        with open(perflog_file_path, "w") as file_:

            if self.nbr_devices:
                file_.write(f"Report for {self.nbr_devices} devices\n")

            total_time = exec_time = int((time() - self.start_time) * 1000)
            file_.write(f"Total execution time: {print_from_ms(total_time)}\n")

            for funct, exec_time in self.times.items():
                if self.nbr_devices:
                    exec_time_per_dev = exec_time / self.nbr_devices
                    log = f"{funct} finished in {print_from_ms(exec_time)} | {print_from_ms(exec_time_per_dev)} per device"

                else:
                    log = f"{funct} finished in {print_from_ms(exec_time)}"

                file_.write(log + "\n")



"""Library of function to convert data from vendor specific format to network_importer.

(c) 2020 Network To Code

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
  http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import logging

from network_importer.utils import is_interface_lag
from network_importer.processors.get_neighbors import Neighbors, Neighbor
from network_importer.processors.get_vlans import Vlans, Vlan

LOGGER = logging.getLogger("network-importer")


def convert_cisco_genie_lldp_neighbors_details(device_name, data):
    """Convert the data returned by Genie for show lldp neighbors detail to Neighbors().

    Args:
        device_name (str): the name of the device where the data was collected
        data (Dict): the parsed data returned by Genie

    Returns:
        Neighbors: List of neighbors in a Pydantic model
    """
    results = Neighbors()

    if "interfaces" not in data:
        return results

    for intf_name, intf_data in data["interfaces"].items():
        if "port_id" not in intf_data.keys():
            continue

        # intf_name = canonical_interface_name(intf_name)
        for nei_intf_name in list(intf_data["port_id"].keys()):

            # nei_intf_name_long = canonical_interface_name(nei_intf_name)
            if is_interface_lag(nei_intf_name):
                LOGGER.debug(
                    "%s | Neighbors, %s is connected to %s but is not a valid interface (lag), SKIPPING",
                    device_name,
                    nei_intf_name,
                    intf_name,
                )
                continue

            if "neighbors" not in intf_data["port_id"][nei_intf_name]:
                LOGGER.debug("%s | No neighbor found for %s connected to %s", device_name, nei_intf_name, intf_name)
                continue

            if len(intf_data["port_id"][nei_intf_name]["neighbors"]) > 1:
                LOGGER.warning(
                    "%s | More than 1 neighbor found for %s connected to %s, SKIPPING",
                    device_name,
                    nei_intf_name,
                    intf_name,
                )
                continue

            neighbor = Neighbor(
                hostname=list(intf_data["port_id"][nei_intf_name]["neighbors"].keys())[0], port=nei_intf_name
            )

            results.neighbors[intf_name].append(neighbor)

    return results


def convert_cisco_genie_cdp_neighbors_details(device_name, data):
    """Convert the data returned by Genie for show cdp neighbors detail to Neighbors().

    Args:
        device_name (str): the name of the device where the data was collected
        data (Dict): the parsed data returned by Genie

    Returns:
        Neighbors: List of neighbors in a Pydantic model
    """
    results = Neighbors()

    if "index" not in data:
        return results

    for _, intf_data in data["index"].items():
        intf_name = intf_data.get("local_interface", None)
        nei_intf_name = intf_data.get("port_id", None)
        neighbor_name = intf_data.get("device_id", None)

        if not intf_name or not nei_intf_name or not neighbor_name:
            continue

        if is_interface_lag(nei_intf_name):
            LOGGER.debug(
                "%s | Neighbors, %s is connected to %s but is not a valid interface (lag), SKIPPING",
                device_name,
                nei_intf_name,
                intf_name,
            )
            continue

        neighbor = Neighbor(hostname=neighbor_name, port=nei_intf_name)
        results.neighbors[intf_name].append(neighbor)

        # Ensure each interface no not have more than 1 neighbor
        for intf_name, value in results.neighbors.items():
            if len(value) == 1:
                continue

            LOGGER.warning(
                "%s | More than 1 neighbor found for %s, SKIPPING",
                device_name,
                intf_name,
            )
            results.neighbors[intf_name] = []

    return results


def convert_cisco_genie_vlans(device_name: str, data: dict) -> Vlans:
    """Convert data returned by Genie from "show vlans" to the Vlans object.

    Args:
        device_name (str): name if the device
        data (dict): Structured Data returned by Genie

    Returns:
        Vlans: List of Vlan in a pydantic model
    """
    results = Vlans()

    if "vlans" not in data:
        return results

    for vid, vlan_data in data["vlans"].items():
        if not vlan_data.get("name", None):
            LOGGER.warning("%s | Unknown VLAN data, VLAN %s", device_name, vid)
            continue

        if vlan_data.get("state", None) == "unsupport":
            LOGGER.debug("%s | Unsupported VLAN found, VLAN %s", device_name, vid)
            continue

        results.vlans.append(Vlan(name=vlan_data["name"], vid=int(vlan_data["vlan_id"])))

    return results



"""GetVlans processor for the network_importer.

(c) 2020 Network To Code

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
  http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import logging
from typing import List

from pydantic import BaseModel  # pylint: disable=no-name-in-module

from network_importer.processors import BaseProcessor

LOGGER = logging.getLogger("network-importer")


# ------------------------------------------------------------
# Standard model to return for get_vlans
# ------------------------------------------------------------
class Vlan(BaseModel):
    """Dataclass model to store one vlan returned by get_vlans."""

    name: str
    vid: int


class Vlans(BaseModel):
    """Dataclass model to store Vlans returned by get_vlans."""

    vlans: List[Vlan] = list()


# ------------------------------------------------------------
# Processor
# ------------------------------------------------------------
class GetVlans(BaseProcessor):
    """Placeholder for GetVlans processor, currently using the BaseProcessor."""



"""Norning Inventory for nautobot."""
# Disable too-many-arguments and too-many-locals pylint tests for this file. These are both necessary
# pylint: disable=R0913,R0914,E1101,W0613

import sys
from typing import Any, List
import pynautobot
from pydantic import ValidationError

from nornir.core.inventory import Defaults, Groups, Hosts, Inventory, ParentGroups, ConnectionOptions
from nornir.core.plugins.inventory import InventoryPluginRegister
from network_importer.inventory import NetworkImporterInventory, NetworkImporterHost
from network_importer.utils import build_filter_params
from network_importer.adapters.nautobot_api.settings import InventorySettings


class NautobotAPIInventory(NetworkImporterInventory):
    """Nautobot API Inventory Class."""

    # pylint: disable=dangerous-default-value, too-many-branches, too-many-statements
    def __init__(
        self,
        *args,
        **kwargs: Any,
    ) -> None:
        """Nornir Inventory Plugin for Nautobot API."""
        super().__init__(
            *args,
            **kwargs,
        )

        try:
            self.settings = InventorySettings(**self.settings)
        except ValidationError as exc:
            print(f"Inventory Settings are not valid, found {len(exc.errors())} error(s)")
            for error in exc.errors():
                print(f"  inventory/{'/'.join(error['loc'])} | {error['msg']} ({error['type']})")
            sys.exit(1)

        # Build Filter based on inventory_settings filter and on limit
        self.filter_parameters = {}
        if self.settings.filter is not None:
            build_filter_params(self.settings.filter.split((",")), self.filter_parameters)

        if self.limit:
            if "=" not in self.limit:
                self.filter_parameters["name"] = self.limit
            else:
                build_filter_params(self.limit.split((",")), self.filter_parameters)

        if "exclude" not in self.filter_parameters.keys():
            self.filter_parameters["exclude"] = "config_context"

        # Instantiate nautobot session using pynautobot
        self.session = pynautobot.api(url=self.settings.address, token=self.settings.token)
        if not self.settings.verify_ssl:
            self.session.http_session.verify = False

    def load(self):
        """Load inventory by fetching devices from nautobot."""
        if self.filter_parameters:
            devices: List[pynautobot.modules.dcim.Devices] = self.session.dcim.devices.filter(**self.filter_parameters)
        else:
            devices: List[pynautobot.modules.dcim.Devices] = self.session.dcim.devices.all()

        # fetch all platforms from nautobot and build mapping:   platform:  napalm_driver
        platforms = self.session.dcim.platforms.all()
        platforms_mapping = {platform.slug: platform.napalm_driver for platform in platforms if platform.napalm_driver}

        hosts = Hosts()
        groups = Groups()
        defaults = Defaults()

        for dev in devices:
            # nautobot allows devices to be unnamed, but the Nornir model does not allow this
            # If a device is unnamed we will set the name to the id of the device in nautobot
            dev_name = dev.name or dev.id
            host = NetworkImporterHost(name=dev_name, connection_options=ConnectionOptions())

            # Only add virtual chassis master as inventory element
            if dev.virtual_chassis and dev.virtual_chassis.master:
                if dev.id != dev.virtual_chassis.master.id:
                    continue
                host.data["virtual_chassis"] = True

            else:
                host.data["virtual_chassis"] = False

            # If supported_platforms is provided
            # skip all devices that do not match the list of supported platforms
            # TODO need to see if we can filter when doing the query directly
            if self.supported_platforms:
                if not dev.platform:
                    continue

                if dev.platform.slug not in self.supported_platforms:
                    continue

            # Add value for IP address
            if self.settings.use_primary_ip and dev.primary_ip:
                host.hostname = dev.primary_ip.address.split("/")[0]
            elif self.settings.use_primary_ip and not dev.primary_ip:
                host.is_reachable = False
                host.not_reachable_reason = "primary ip not defined in nautobot"
            elif not self.settings.use_primary_ip and self.settings.fqdn:
                host.hostname = f"{dev.name}.{self.settings.fqdn}"
            elif not self.settings.use_primary_ip:
                host.hostname = dev.name
            else:
                host.hostname = dev_name

            host.site_name = dev.site.slug

            host.data["serial"] = dev.serial
            host.data["vendor"] = dev.device_type.manufacturer.slug
            host.data["asset_tag"] = dev.asset_tag
            host.data["custom_fields"] = dev.custom_fields
            host.data["site_id"] = dev.site.id
            host.data["device_id"] = dev.id
            host.data["role"] = dev.device_role.slug
            host.data["model"] = dev.device_type.slug

            # Attempt to add 'platform' based of value in 'slug'
            if dev.platform and dev.platform.slug in platforms_mapping:
                host.connection_options = {"napalm": ConnectionOptions(platform=platforms_mapping[dev.platform.slug])}

            if dev.platform:
                host.platform = dev.platform.slug
            else:
                host.platform = None

            host.groups = ParentGroups([self.global_group])

            if dev.site.slug not in groups.keys():
                groups[dev.site.slug] = {}

            if dev.device_role.slug not in groups.keys():
                groups[dev.device_role.slug] = {}

            if host.hostname and host.platform:
                host.is_reachable = True

            # Assign temporary dict to outer dict

            hosts[dev_name] = host

        return Inventory(hosts=hosts, groups=groups, defaults=defaults)


InventoryPluginRegister.register("NautobotAPIInventory", NautobotAPIInventory)



"""Nornir Inventory for netbox."""

# pylint: disable=no-member,too-many-arguments,unused-argument,too-many-branches,too-many-statements

import sys
from typing import Any, List
import requests
import pynetbox
from pydantic import ValidationError

from nornir.core.inventory import Defaults, Groups, Hosts, Inventory, ParentGroups, ConnectionOptions
from nornir.core.plugins.inventory import InventoryPluginRegister
from network_importer.inventory import NetworkImporterInventory, NetworkImporterHost
from network_importer.utils import build_filter_params

from network_importer.adapters.netbox_api.settings import InventorySettings


class NetBoxAPIInventory(NetworkImporterInventory):
    """Netbox API Inventory Class."""

    def __init__(
        self,
        *args,
        **kwargs: Any,
    ) -> None:
        """Nornir Inventory Plugin for Netbox API."""
        super().__init__(*args, **kwargs)

        try:
            self.settings = InventorySettings(**self.settings)
        except ValidationError as exc:
            print(f"Inventory Settings are not valid, found {len(exc.errors())} error(s)")
            for error in exc.errors():
                print(f"  inventory/{'/'.join(error['loc'])} | {error['msg']} ({error['type']})")
            sys.exit(1)

        # Build Filter based on inventory_settings filter and on limit
        self.filter_parameters = {}
        if self.settings.filter is not None:
            build_filter_params(self.settings.filter.split((",")), self.filter_parameters)

        if self.limit:
            if "=" not in self.limit:
                self.filter_parameters["name"] = self.limit
            else:
                build_filter_params(self.limit.split((",")), self.filter_parameters)

        if "exclude" not in self.filter_parameters.keys():
            self.filter_parameters["exclude"] = "config_context"

        # Instantiate netbox session using pynetbox
        self.session = pynetbox.api(url=self.settings.address, token=self.settings.token)
        if not self.settings.verify_ssl:
            session = requests.Session()
            session.verify = False
            self.session.http_session = session

    def load(self):
        """Load inventory by fetching devices from netbox."""
        if self.filter_parameters:
            devices: List[pynetbox.modules.dcim.Devices] = self.session.dcim.devices.filter(**self.filter_parameters)
        else:
            devices: List[pynetbox.modules.dcim.Devices] = self.session.dcim.devices.all()

        # fetch all platforms from Netbox and build mapping:   platform:  napalm_driver
        platforms = self.session.dcim.platforms.all()
        platforms_mapping = {platform.slug: platform.napalm_driver for platform in platforms if platform.napalm_driver}

        hosts = Hosts()
        groups = Groups()
        defaults = Defaults()

        for dev in devices:
            # Netbox allows devices to be unnamed, but the Nornir model does not allow this
            # If a device is unnamed we will set the name to the id of the device in netbox
            dev_name = dev.name or dev.id
            host = NetworkImporterHost(name=dev_name, connection_options=ConnectionOptions())

            # Only add virtual chassis master as inventory element
            if dev.virtual_chassis and dev.virtual_chassis.master:
                if dev.id != dev.virtual_chassis.master.id:
                    continue
                host.data["virtual_chassis"] = True

            else:
                host.data["virtual_chassis"] = False

            # If supported_platforms is provided
            # skip all devices that do not match the list of supported platforms
            # TODO need to see if we can filter when doing the query directly
            if self.supported_platforms:
                if not dev.platform:
                    continue

                if dev.platform.slug not in self.supported_platforms:
                    continue

            # Add value for IP address
            if self.settings.use_primary_ip and dev.primary_ip:
                host.hostname = dev.primary_ip.address.split("/")[0]
            elif self.settings.use_primary_ip and not dev.primary_ip:
                host.is_reachable = False
                host.not_reachable_reason = "primary ip not defined in Netbox"
            elif not self.settings.use_primary_ip and self.settings.fqdn:
                host.hostname = f"{dev.name}.{self.settings.fqdn}"
            elif not self.settings.use_primary_ip:
                host.hostname = dev.name
            else:
                host.hostname = dev_name

            host.site_name = dev.site.slug

            host.data["serial"] = dev.serial
            host.data["vendor"] = dev.device_type.manufacturer.slug
            host.data["asset_tag"] = dev.asset_tag
            host.data["custom_fields"] = dev.custom_fields
            host.data["site_id"] = dev.site.id
            host.data["device_id"] = dev.id
            host.data["role"] = dev.device_role.slug
            host.data["model"] = dev.device_type.slug

            # Attempt to add 'platform' based of value in 'slug'
            if dev.platform and dev.platform.slug in platforms_mapping:
                host.connection_options = {"napalm": ConnectionOptions(platform=platforms_mapping[dev.platform.slug])}

            if dev.platform:
                host.platform = dev.platform.slug
            else:
                host.platform = None

            host.groups = ParentGroups([self.global_group])

            if dev.site.slug not in groups.keys():
                groups[dev.site.slug] = {}

            if dev.device_role.slug not in groups.keys():
                groups[dev.device_role.slug] = {}

            if host.hostname and host.platform:
                host.is_reachable = True

            # Assign temporary dict to outer dict

            hosts[dev_name] = host

        return Inventory(hosts=hosts, groups=groups, defaults=defaults)


InventoryPluginRegister.register("NetBoxAPIInventory", NetBoxAPIInventory)



"""Custom Exceptions for the NetworkImporterAdapter."""
import re
import ipaddress
import json
import logging
from collections import defaultdict

from diffsync.exceptions import ObjectNotFound, ObjectAlreadyExists

from pybatfish.client.session import Session
from pybatfish.exception import BatfishException

# pylint: disable=import-error
import network_importer.config as config
from network_importer.adapters.base import BaseAdapter

from network_importer.exceptions import AdapterLoadFatalError
from network_importer.inventory import reachable_devs, valid_and_reachable_devs
from network_importer.tasks import check_if_reachable, warning_not_reachable
from network_importer.drivers import dispatcher
from network_importer.processors.get_neighbors import GetNeighbors, hosts_for_cabling
from network_importer.processors.get_vlans import GetVlans
from network_importer.utils import (
    is_interface_lag,
    is_interface_physical,
    expand_vlans_list,
)
from network_importer.adapters.network_importer.exceptions import BatfishObjectNotValid

LOGGER = logging.getLogger("network-importer")


class NetworkImporterAdapter(BaseAdapter):
    """Adapter to import data from a network based on Batfish."""

    top_level = ["site", "device", "cable"]

    type = "Network"

    bfi = None

    def load(self):
        """Initialize batfish and load all data from the network in the local cache."""
        sites = {}

        self.init_batfish()

        # Create all devices and site object from Nornir Inventory
        for hostname, host in self.nornir.inventory.hosts.items():

            if len(self.bfi.q.nodeProperties(nodes=f'"{hostname}"').answer()) == 0:
                self.nornir.inventory.hosts[hostname].has_config = False
                LOGGER.warning("Unable to find information for %s in Batfish, SKIPPING", hostname)
                continue

            self.nornir.inventory.hosts[hostname].has_config = True

            if host.site_name not in sites.keys():
                site = self.site(name=host.site_name)
                sites[host.site_name] = site
                self.add(site)
            else:
                site = sites[host.site_name]

            device = self.device(name=hostname, site_name=host.site_name)
            self.add(device)

        if config.SETTINGS.main.import_cabling in ["lldp", "cdp"] or config.SETTINGS.main.import_vlans in [True, "cli"]:
            self.nornir.filter(filter_func=reachable_devs).run(task=check_if_reachable, on_failed=True)
            self.nornir.filter(filter_func=reachable_devs).run(task=warning_not_reachable, on_failed=True)

        self.load_batfish()
        self.load_vlans()
        self.load_cabling()

        self.check_data_consistency()

    def init_batfish(self):
        """Initialize Batfish snapshot and session."""
        network_name = config.SETTINGS.batfish.network_name
        snapshot_name = config.SETTINGS.batfish.snapshot_name
        snapshot_path = config.SETTINGS.main.configs_directory

        bf_params = dict(
            host=config.SETTINGS.batfish.address,
            port_v1=config.SETTINGS.batfish.port_v1,
            port_v2=config.SETTINGS.batfish.port_v2,
            ssl=config.SETTINGS.batfish.use_ssl,
        )
        if config.SETTINGS.batfish.api_key:
            bf_params["api_key"] = config.SETTINGS.batfish.api_key

        try:
            self.bfi = Session.get("bf", **bf_params)
            self.bfi.verify = False
            self.bfi.set_network(network_name)
            self.bfi.init_snapshot(snapshot_path, name=snapshot_name, overwrite=True)
        except BatfishException as exc:
            error = json.loads(str(exc).splitlines()[-1])
            error = re.sub(r"[^:]*:.", "", error["answerElements"][0]["answer"][0])
            raise AdapterLoadFatalError(error) from exc

    def load_batfish(self):
        """Load all devices, interfaces and IP Addresses from Batfish."""
        # Import Devices
        devices = self.get_all(self.device)

        for device in devices:
            self.load_batfish_device(device=device)

    def load_batfish_device(self, device):
        """Load all devices from Batfish.

        Args:
            device (Device): Device object
        """
        site = self.get(self.site, identifier=device.site_name)

        interface_vlans_mapping = defaultdict(list)
        if config.SETTINGS.main.import_vlans:
            bf_vlans = self.bfi.q.switchedVlanProperties(nodes=f'"{device.name}"').answer()
            for bf_vlan in bf_vlans.frame().itertuples():

                if config.SETTINGS.main.import_vlans in ["config", True]:
                    vlan, created = self.get_or_add(
                        self.vlan(name=f"vlan-{bf_vlan.VLAN_ID}", vid=bf_vlan.VLAN_ID, site_name=site.name)
                    )
                    if created:
                        site.add_child(vlan)

                    vlan.add_device(device.get_unique_id())

                # Save interface to vlan mapping for later use
                for intf in bf_vlan.Interfaces:
                    if intf.hostname != device.name.lower():
                        continue
                    interface_vlans_mapping[intf.interface].append(
                        self.vlan.create_unique_id(vid=bf_vlan.VLAN_ID, site_name=site.name)
                    )

        intfs = self.bfi.q.interfaceProperties(nodes=f'"{device.name}"').answer().frame()
        for _, intf in intfs.iterrows():
            self.load_batfish_interface(
                site=site,
                device=device,
                intf=dict(intf),
                interface_vlans=interface_vlans_mapping[intf["Interface"].interface],
            )

    def load_batfish_interface(
        self, site, device, intf, interface_vlans=[]
    ):  # pylint: disable=dangerous-default-value,unused-argument,too-many-statements,too-many-branches,too-many-locals
        """Load an interface for a given device from Batfish Data, including IP addresses and prefixes.

        Args:
            site (Site): Site object
            device (Device): Device object
            intf (dict): Batfish interface object in Dict format
            interface_vlans (list[str], optional): List of vlan uis associated with this interface

        Returns:
            Interface
        """
        # try:
        #     self._check_batfish_interface_is_valid(intf)
        # except BatfishObjectNotValid as exc:
        #     LOGGER.warning("Unable to add an interface on %s, the data is not valid (%s)", device.name, str(exc))
        #     return False

        # Since it's possible to import vlans from the config and or from the CLI, we need to track 2 differents flags
        #  import_vlans = import the vlans information to the interface, populate allowed_vlans, access_vlan and mode
        #  create_vlans = create the vlans object
        import_vlans = create_vlans = False
        if config.SETTINGS.main.import_vlans in ["config", True, "yes"]:
            create_vlans = True
        if config.SETTINGS.main.import_vlans not in [False, "no"]:
            import_vlans = True

        interface = self.interface(
            name=intf["Interface"].interface,
            device_name=device.name,
            mtu=intf["MTU"],
            switchport_mode=intf["Switchport_Mode"],
        )

        if "Description" in intf and intf["Description"]:
            interface.description = intf["Description"].strip()

        is_physical = is_interface_physical(interface.name)
        is_lag = is_interface_lag(interface.name)

        if is_lag:
            interface.is_lag = True
            interface.is_virtual = False
        elif is_physical is False:  # pylint: disable=C0121
            interface.is_virtual = True
        else:
            interface.is_virtual = False

        if interface.switchport_mode == "FEX_FABRIC":
            interface.switchport_mode = "NONE"

        if config.SETTINGS.main.import_intf_status:
            interface.active = intf["Active"]
        elif not config.SETTINGS.main.import_intf_status:
            interface.active = None

        if interface.is_lag is None and interface.lag_members is None and len(list(intf["Channel_Group_Members"])) != 0:
            interface.lag_members = list(intf["Channel_Group_Members"])
            interface.is_lag = True
            interface.is_virtual = False
        elif interface.is_lag is None:
            interface.is_lag = False

        if interface.mode is None and interface.switchport_mode:
            if intf["Encapsulation_VLAN"]:
                interface.mode = "L3_SUB_VLAN"
                vlan = self.vlan(vid=intf["Encapsulation_VLAN"], site_name=site.name)
                vlan, _ = self.get_or_create_vlan(vlan, site)
                if import_vlans:
                    interface.allowed_vlans = [vlan.get_unique_id()]
                    interface_vlans = interface.allowed_vlans
            else:
                interface.mode = interface.switchport_mode

        if interface.mode == "TRUNK":
            vids = expand_vlans_list(intf["Allowed_VLANs"])
            for vid in vids:
                vlan = self.vlan(vid=vid, site_name=site.name)
                if create_vlans:
                    vlan, _ = self.get_or_create_vlan(vlan, site)
                if import_vlans:
                    interface.allowed_vlans.append(vlan.get_unique_id())

            interface_vlans = interface.allowed_vlans

            if intf["Native_VLAN"]:
                native_vlan = self.vlan(vid=intf["Native_VLAN"], site_name=site.name)
                if create_vlans:
                    native_vlan, _ = self.get_or_create_vlan(native_vlan)
                if import_vlans:
                    interface.access_vlan = native_vlan.get_unique_id()
                    interface_vlans = [interface.access_vlan]

        elif interface.mode == "ACCESS" and intf["Access_VLAN"]:
            vlan = self.vlan(vid=intf["Access_VLAN"], site_name=site.name)
            if create_vlans:
                vlan, _ = self.get_or_create_vlan(vlan, site)
            if import_vlans:
                interface.access_vlan = vlan.get_unique_id()
                interface_vlans = [interface.access_vlan]

        if interface.is_lag is False and interface.is_lag_member is None and intf["Channel_Group"]:
            interface.parent = self.interface(name=intf["Channel_Group"], device_name=device.name).get_unique_id()
            interface.is_lag_member = True
            interface.is_virtual = False

        self.add(interface)
        device.add_child(interface)

        if "All_Prefixes" not in intf:
            return interface

        for prefix in intf["All_Prefixes"]:
            self.load_batfish_ip_address(
                site=site, device=device, interface=interface, address=prefix, interface_vlans=interface_vlans
            )

        return interface

    def load_batfish_ip_address(
        self, site, device, interface, address, interface_vlans=[]
    ):  # pylint: disable=dangerous-default-value,too-many-arguments
        """Load IP address for a given interface from Batfish.

        Args:
            site (Site): Site object
            device (Device): Device object
            interface (Interface): Interface object
            address (str): IP address in string format
            interface_vlans (list[str], optional): List of vlan uis associated with this interface

        Returns:
            IPAddress
        """
        ip_address = self.ip_address(
            address=address,
            device_name=device.name,
            interface_name=interface.name,
        )

        if config.SETTINGS.main.import_ips:
            LOGGER.debug("%s | Import %s for %s::%s", self.name, ip_address.address, device.name, interface.name)
            try:
                self.add(ip_address)
                interface.add_child(ip_address)
            except ObjectAlreadyExists:
                LOGGER.error(
                    "%s | Duplicate IP found for %s (%s) ; IP already imported.", self.name, ip_address, device.name
                )

        if config.SETTINGS.main.import_prefixes:
            vlan = None
            if len(interface_vlans) == 1:
                vlan = interface_vlans[0]
            elif len(interface_vlans) >= 1:
                LOGGER.warning(
                    "%s | More than 1 vlan associated with interface %s (%s)",
                    device.name,
                    interface.name,
                    interface_vlans,
                )

            self.add_prefix_from_ip(ip_address=ip_address, site=site, vlan=vlan)

        return ip_address

    def add_prefix_from_ip(self, ip_address, site, vlan=None):
        """Try to extract a prefix from an IP address and save it locally.

        Args:
            ip_address (IPAddress): DiffSync IPAddress object
            site (Site): Site object the prefix is part of.
            vlan (str): Identifier of the vlan

        Returns:
            Prefix, bool: False if a prefix can't be extracted from this IP address
        """
        prefix = ipaddress.ip_network(ip_address.address, strict=False)

        if prefix.num_addresses == 1:
            return False

        try:
            prefix_obj = self.get(self.prefix, identifier=dict(site_name=site.name, prefix=prefix))
        except ObjectNotFound:
            prefix_obj = None

        if not prefix_obj:
            prefix_obj = self.prefix(prefix=str(prefix), site_name=site.name, vlan=vlan)
            self.add(prefix_obj)
            site.add_child(prefix_obj)
            LOGGER.debug("Added Prefix %s from batfish", prefix)

        if prefix_obj and vlan and not prefix_obj.vlan:
            prefix_obj.vlan = vlan
            LOGGER.debug("Updated Prefix %s with vlan %s", prefix, vlan)

        return prefix_obj

    def load_cabling(self):
        """Load cabling from either batfish, cdl or lldp based on the configuration."""
        if config.SETTINGS.main.import_cabling in ["no", False]:
            return False

        if config.SETTINGS.main.import_cabling in ["config", True]:
            self.load_batfish_cable()

        if config.SETTINGS.main.import_cabling in ["lldp", "cdp", True]:
            self.load_cabling_from_cmds()

        self.validate_cabling()

        return True

    def load_vlans(self):
        """Load vlans information from the devices using CLI."""
        if config.SETTINGS.main.import_vlans not in ["cli", True]:
            return

        LOGGER.info("Collecting vlans information from devices .. ")

        results = (
            self.nornir.filter(filter_func=valid_and_reachable_devs)
            .with_processors([GetVlans()])
            .run(task=dispatcher, method="get_vlans")
        )

        for dev_name, items in results.items():
            if items[0].failed:
                continue

            if not isinstance(items[1].result, dict) or "vlans" not in items[1].result:
                LOGGER.debug("%s | No vlan information returned SKIPPING", dev_name)
                continue

            device = self.get(self.device, identifier=dev_name)
            site = self.get(self.site, identifier=device.site_name)

            for vlan in items[1].result["vlans"]:
                new_vlan, created = self.get_or_add(self.vlan(vid=vlan["vid"], name=vlan["name"], site_name=site.name))

                if created:
                    site.add_child(new_vlan)

                new_vlan.add_device(device.get_unique_id())

    def load_batfish_cable(self):
        """Import cables from Batfish using layer3Edges tables."""
        device_names = [device.name for device in self.get_all(self.device)]
        p2p_links = self.bfi.q.layer3Edges().answer()
        existing_cables = []
        for link in p2p_links.frame().itertuples():

            if link.Interface.hostname not in device_names:
                continue

            if link.Remote_Interface.hostname not in device_names:
                continue

            cable = self.cable(
                device_a_name=link.Interface.hostname,
                interface_a_name=re.sub(r"\.\d+$", "", link.Interface.interface),
                device_z_name=link.Remote_Interface.hostname,
                interface_z_name=re.sub(r"\.\d+$", "", link.Remote_Interface.interface),
                source="batfish",
            )
            uid = cable.get_unique_id()

            if uid not in existing_cables:
                self.add(cable)
                existing_cables.append(uid)

        nbr_cables = len(self.get_all(self.cable))
        LOGGER.debug("Found %s cables in Batfish", nbr_cables)

    def load_cabling_from_cmds(self):
        """Import cabling information from the CLI, either using LDLP or CDP based on the configuration.

        If the FQDN is defined, and the hostname of a neighbor include the FQDN, remove it.
        """
        LOGGER.info("Collecting cabling information from devices .. ")

        results = (
            self.nornir.filter(filter_func=valid_and_reachable_devs)
            .filter(filter_func=hosts_for_cabling)
            .with_processors([GetNeighbors()])
            .run(
                task=dispatcher,
                method="get_neighbors",
                on_failed=True,
            )
        )

        nbr_cables = 0
        for dev_name, items in results.items():
            if items[0].failed:
                continue

            if not isinstance(items[1][0].result, dict) or "neighbors" not in items[1][0].result:
                LOGGER.debug("%s | No neighbors information returned SKIPPING", dev_name)
                continue

            for interface, neighbors in items[1][0].result["neighbors"].items():
                cable = self.cable(
                    device_a_name=dev_name,
                    interface_a_name=interface,
                    device_z_name=neighbors[0]["hostname"],
                    interface_z_name=neighbors[0]["port"],
                    source="cli",
                )
                nbr_cables += 1
                LOGGER.debug("%s | Added cable %s", dev_name, cable.get_unique_id())
                self.get_or_add(cable)

        LOGGER.debug("Found %s cables from Cli", nbr_cables)

    def check_data_consistency(self):
        """Check the validaty and consistency of the data in the local store.

        Ensure the vlans configured for each interface exist in the system.
        On some vendors, it's possible to have a list larger than what is really available

        In the process, ensure that all devices are associated with the right vlans
        """
        # Get a dictionnary with all vlans organized by uid
        vlans = {vlan.get_unique_id(): vlan for vlan in self.get_all(self.vlan)}
        interfaces = self.get_all(self.interface)

        for intf in interfaces:
            if intf.allowed_vlans:
                clean_allowed_vlans = []
                for vlan in intf.allowed_vlans:
                    if vlan in vlans.keys():
                        clean_allowed_vlans.append(vlan)
                        vlans[vlan].add_device(intf.device_name)

                intf.allowed_vlans = clean_allowed_vlans

    def validate_cabling(self):
        """Check if all cables are valid.

        Check if all cables are valid
        Check if both devices are present in the device list
            For now only process cables with both devices present
        Check if both interfaces are present as well and are not virtual
        Check that both interfaces are not already connected to a different device/interface

        When a cable is not valid, update the flag valid on the object itself
        Non valid cables will be ignored later on for update/creation
        """

        def is_cable_side_valid(cable, side):
            """Check if the given side of a cable (a or z) is valid or not.

            Check if both the device and the interface are present in the internal store
            """
            dev_name, intf_name = cable.get_device_intf(side)

            try:
                self.get(self.device, identifier=dev_name)
            except ObjectNotFound:
                return True

            try:
                intf = self.get(self.interface, identifier=dict(name=intf_name, device_name=dev_name))
            except ObjectNotFound:
                return True

            # if not dev:
            #     LOGGER.debug("CABLE: %s not present in devices list (%s side)", dev_name, side)
            #     self.delete(cable)
            #     return False

            # intf = self.get(self.interface, keys=[dev_name, intf_name])

            # if not intf:
            #     LOGGER.warning("CABLE: %s:%s not present in interfaces list", dev_name, intf_name)
            #     self.delete(cable)
            #     return False

            if intf.is_virtual:
                LOGGER.debug(
                    "CABLE: %s:%s is a virtual interface, can't be used for cabling SKIPPING (%s side)",
                    dev_name,
                    intf_name,
                    side,
                )
                self.remove(cable)
                return False

            return True

        cables = self.get_all(self.cable)
        for cable in list(cables):

            for side in ["a", "z"]:

                if not is_cable_side_valid(cable, side):
                    break

                # remote_side = "z"
                # if side == "z":
                #     remote_side = "a"

                # remote_device_expected, remote_intf_expected = cable.get_device_intf(remote_side)

                # remote_dev = self.get(self.device, keys=[remote_device_expected])

                # if not dev.interfaces[intf_name].remote or not dev.interfaces[intf_name].remote.remote:
                #     continue

                # cable_type = dev.interfaces[intf_name].remote.remote.connected_endpoint_type

                # # Check if the interface is already connected
                # # Check if it's already connected to the right device
                # if not cable_type:
                #     # Interface is currently not connected in netbox
                #     continue

                # elif cable_type != "dcim.interface":

                #     LOGGER.debug(
                #         f"CABLE: {dev_name}:{intf_name} is already connected but to a different type of interface  ({cable_type})"
                #     )
                #     cable.is_valid = False
                #     cable.error = "wrong-cable-type"
                #     continue

                # elif cable_type == "dcim.interface":
                #     remote_host_reported = dev.interfaces[intf_name].remote.remote.connected_endpoint.device.name
                #     remote_int_reported = dev.interfaces[intf_name].remote.remote.connected_endpoint.name

                #     if remote_host_reported != remote_device_expected:
                #         LOGGER.warning(
                #             f"CABLE: {dev_name}:{intf_name} is already connected but to a different device ({remote_host_reported} vs {remote_device_expected})"
                #         )
                #         cable.is_valid = False
                #         cable.error = "wrong-peer-device"
                #         continue

                #     elif remote_host_reported == remote_device_expected and remote_intf_expected != remote_int_reported:
                #         LOGGER.warning(
                #             f"CABLE:  {dev_name}:{intf_name} is already connected but to a different interface ({remote_int_reported} vs {remote_intf_expected})"
                #         )
                #         cable.is_valid = False
                #         cable.error = "interface-already-connected"
                #         continue

    @staticmethod
    def _check_batfish_interface_is_valid(intf):
        """Check if a given dict meant to represent a batfish interface is valid.

        Args:
            intf (dict): dict generated from a Batfish Interface
        """
        if not isinstance(intf, dict):
            raise BatfishObjectNotValid("batfish interface is not a dict")

        mandatory_fields = ["Interface", "Description", "MTU", "Switchport_Mode"]

        for field in mandatory_fields:
            if field not in intf.keys():
                raise BatfishObjectNotValid(f"Key {field} is missing in batfish interface")

        return True



"""Main dispatcher for nornir."""
import logging
import importlib

# from nornir.core.exceptions import NornirSubTaskError
from nornir.core.task import Result, Task

import network_importer.config as config

LOGGER = logging.getLogger("network-importer")


def dispatcher(task: Task, method: str) -> Result:
    """Helper Task to retrieve a given Nornir task for a given platform
    Args:
        task (Nornir Task):  Nornir Task object
        task (Nornir Task):  Nornir Task object
    Returns:
        Result: Nornir Task result
    """

    LOGGER.debug("Executing dispatcher for %s (%s)", task.host.name, task.host.platform)

    # Get the platform specific driver, if not available, get the default driver
    driver = config.SETTINGS.drivers.mapping.get(task.host.platform, config.SETTINGS.drivers.mapping.get("default"))
    LOGGER.debug("Found driver %s", driver)

    if not driver:
        LOGGER.warning(
            "%s | Unable to find the driver for %s for platform : %s", task.host.name, method, task.host.platform
        )
        return Result(host=task.host, failed=True)

    driver_class = getattr(importlib.import_module(driver), "NetworkImporterDriver")

    if not driver_class:
        LOGGER.error("%s | Unable to locate the class %s", task.host.name, driver)
        return Result(host=task.host, failed=True)

    try:
        driver_task = getattr(driver_class, method)
    except AttributeError:
        LOGGER.error("%s | Unable to locate the method %s for %s", task.host.name, method, driver)
        return Result(host=task.host, failed=True)

    result = task.run(task=driver_task)

    return Result(host=task.host, result=result)



"""Settings definition for the NetboxAPIAdapter."""
from typing import List, Optional, Union
from pydantic import BaseSettings

from diffsync import DiffSyncModelFlags


class AdapterSettings(BaseSettings):
    """Config settings for the netbox_api adapter. Not used currently."""

    model_flag_tags: List[str] = list()  # List of tags that defines what objects to assign the model_flag to.
    model_flag: Optional[DiffSyncModelFlags]  # The model flag that will be applied to objects based on tag.


class InventorySettings(BaseSettings):
    """Config settings for the NautobotAPI inventory."""

    address: Optional[str] = "http://localhost"
    token: Optional[str] = None
    verify_ssl: Union[bool, str] = True

    use_primary_ip: Optional[bool] = True
    fqdn: Optional[str] = None
    filter: Optional[str] = None

    class Config:
        """Additional parameters to automatically map environment variable to some settings."""

        fields = {
            "address": {"env": "NAUTOBOT_ADDRESS"},
            "token": {"env": "NAUTOBOT_TOKEN"},
            "verify_ssl": {"env": "NAUTOBOT_VERIFY_SSL"},
        }



"""default network_importer driver for cisco.

(c) 2020 Network To Code

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
  http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import logging

from nornir_netmiko.tasks import netmiko_send_command
from nornir.core.task import Result, Task
from nornir.core.exceptions import NornirSubTaskError

from netmiko.ssh_exception import (
    NetmikoAuthenticationException,
    NetmikoTimeoutException,
)

import network_importer.config as config
from network_importer.drivers.default import NetworkImporterDriver as DefaultNetworkImporterDriver
from network_importer.drivers.converters import (
    convert_cisco_genie_lldp_neighbors_details,
    convert_cisco_genie_cdp_neighbors_details,
    convert_cisco_genie_vlans,
)

LOGGER = logging.getLogger("network-importer")


class NetworkImporterDriver(DefaultNetworkImporterDriver):
    """Collection of Nornir Tasks specific to Cisco devices."""

    @staticmethod
    def get_config(task: Task) -> Result:
        """Get the latest configuration from the device using Netmiko.

        Args:
            task (Task): Nornir Task

        Returns:
            Result: Nornir Result object with a dict as a result containing the running configuration
                { "config: <running configuration> }
        """
        LOGGER.debug("Executing get_config for %s (%s)", task.host.name, task.host.platform)

        try:
            result = task.run(task=netmiko_send_command, command_string="show run", enable=True)
        except NornirSubTaskError as exc:
            if isinstance(exc.result.exception, NetmikoAuthenticationException):
                LOGGER.warning("Unable get the configuration because it can't connect to %s", task.host.name)
                return Result(host=task.host, failed=True, exception=exc.result.exception)

            if isinstance(exc.result.exception, NetmikoTimeoutException):
                LOGGER.warning("Unable get the configuration because the connection to %s timeout", task.host.name)
                return Result(host=task.host, failed=True, exception=exc.result.exception)

            LOGGER.debug("An exception occurred while pulling the configuration", exc_info=True)
            return Result(host=task.host, failed=True, exception=exc.result.exception)

        if result[0].failed:
            return result

        running_config = result[0].result

        if "ERROR: % Invalid input detected at" in running_config:
            LOGGER.warning("Unable to get the configuration for %s", task.host.name)
            return Result(host=task.host, failed=True)

        return Result(host=task.host, result={"config": running_config})

    @staticmethod
    def get_neighbors(task: Task) -> Result:
        """Get a list of neighbors from the device.

        Args:
            task (Task): Nornir Task

        Returns:
            Result: Nornir Result object with a dict as a result containing the neighbors
            The format of the result but must be similar to Neighbors defined in network_importer.processors.get_neighbors
        """
        LOGGER.debug("Executing get_neighbor for %s (%s)", task.host.name, task.host.platform)

        if config.SETTINGS.main.import_cabling == "lldp":
            command = "show lldp neighbors detail"
            converter = convert_cisco_genie_lldp_neighbors_details
            cmd_type = "LLDP"
        elif config.SETTINGS.main.import_cabling == "cdp":
            command = "show cdp neighbors detail"
            converter = convert_cisco_genie_cdp_neighbors_details
            cmd_type = "CDP"
        else:
            return Result(host=task.host, failed=True)

        try:
            result = task.run(task=netmiko_send_command, command_string=command, use_genie=True)
        except NornirSubTaskError:
            LOGGER.debug("An exception occured while pulling %s data", cmd_type, exc_info=True)
            return Result(host=task.host, failed=True)

        if result[0].failed:
            return result

        results = converter(device_name=task.host.name, data=result[0].result)
        return Result(host=task.host, result=results.dict())

    @staticmethod
    def get_vlans(task: Task) -> Result:
        """Get a list of vlans from the device using netmiko and genie parser.

        Args:
            task (Task): Nornir Task

        Returns:
            Result: Nornir Result object with a dict as a result containing the vlans
            The format of the result but must be similar to Vlans defined in network_importer.processors.get_vlans
        """
        LOGGER.debug("Executing get_vlans for %s (%s)", task.host.name, task.host.platform)

        try:
            results = task.run(task=netmiko_send_command, command_string="show vlan", use_genie=True)
        except NornirSubTaskError:
            LOGGER.debug(
                "An exception occured while pulling the vlans information",
                exc_info=True,
            )
            return Result(host=task.host, failed=True)

        if not isinstance(results[0].result, dict) or "vlans" not in results[0].result:
            LOGGER.warning("%s | No vlans information returned", task.host.name)
            return Result(host=task.host, result=False)

        results = convert_cisco_genie_vlans(device_name=task.host.name, data=results[0].result)
        return Result(host=task.host, result=results.dict())



"""Custom Exceptions for the NetworkImporterAdapter."""


class AdapterLoadFatalError(Exception):
    """Network Importer Exception."""


class ConfigLoadFatalError(Exception):
    """Network Importer Exception."""



"""Collection of nornir tasks for the NetboxAPIAdapter."""
import logging

import pynautobot
from nornir.core.task import Result, Task

import network_importer.config as config  # pylint: disable=import-error
from network_importer.adapters.nautobot_api.settings import InventorySettings

LOGGER = logging.getLogger("network-importer")


def query_device_info_from_nautobot(task: Task) -> Result:
    """Nornir Task to query the device information from Nautobot.

    Currently this task will pull both the device information but th goal is to pull additional information
    and return everything in a single dict
    TODO add logic to pull interfaces as well
    TODO add logic to pull ip addresses as well

    Args:
        task (Task): Nornir Task with a valid network device

    Returns:
        Result: Nornir Result object with the result in a dict format
    """
    inventory_settings = InventorySettings(**config.SETTINGS.inventory.settings)
    nautobot = pynautobot.api(url=inventory_settings.address, token=inventory_settings.token)

    # Check for SSL Verification, set it to false if not. Else set to true
    if not inventory_settings.verify_ssl:
        # No manual session is required for this, pynautobot will automatically create one
        nautobot.http_session.verify = False
    else:
        nautobot.http_session.verify = True

    # Set a Results dictionary
    results = {
        "device": None,
        "interfaces": None,
    }

    # Get the device based on the filter
    device = nautobot.dcim.devices.filter(name=task.host.name)

    # Return a failed that there were too many devices returned in the filterset
    if len(device) > 1:
        LOGGER.warning("More than 1 device returned from Nautobot for %s", task.host.name)
        return Result(host=task.host, failed=True)

    # Return a failed when no devices were returned
    if not device:
        LOGGER.warning("No device returned from Nautobot for %s", task.host.name)
        return Result(host=task.host, failed=True)

    results["device"] = dict(device[0])

    # TODO move the logic to pull the interface and potentially IP here
    # interfaces = netbox.dcim.interfaces.filter(device=task.host.name)
    # results["interfaces"] = [ dict(intf) for intf in interfaces ]
    return Result(host=task.host, result=results)



"""default driver for the network_importer.

(c) 2020 Network To Code

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
  http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import logging

from nornir_napalm.plugins.tasks import napalm_get
from nornir_netmiko.tasks import netmiko_send_command

from nornir.core.task import Result, Task
from nornir.core.exceptions import NornirSubTaskError

import network_importer.config as config
from network_importer.drivers.converters import convert_cisco_genie_cdp_neighbors_details

LOGGER = logging.getLogger("network-importer")


class NetworkImporterDriver:
    """Default collection of Nornir Tasks based on Napalm."""

    @staticmethod
    def get_config(task: Task) -> Result:
        """Get the latest configuration from the device.

        Args:
            task (Task): Nornir Task

        Returns:
            Result: Nornir Result object with a dict as a result containing the running configuration
                { "config: <running configuration> }
        """
        LOGGER.debug("Executing get_config for %s (%s)", task.host.name, task.host.platform)

        try:
            result = task.run(task=napalm_get, getters=["config"], retrieve="running")
        except:  # noqa: E722 # pylint: disable=bare-except
            LOGGER.debug("An exception occured while pulling the configuration", exc_info=True)
            return Result(host=task.host, failed=True)

        if result[0].failed:
            return result

        running_config = result[0].result.get("config", {}).get("running", None)
        return Result(host=task.host, result={"config": running_config})

    @staticmethod
    def get_neighbors(task: Task) -> Result:  # pylint: disable=too-many-return-statements
        """Get a list of neighbors from the device.

        Args:
            task (Task): Nornir Task

        Returns:
            Result: Nornir Result object with a dict as a result containing the neighbors
            The format of the result but must be similar to Neighbors defined in network_importer.processors.get_neighbors
        """
        LOGGER.debug("Executing get_neighbor for %s (%s)", task.host.name, task.host.platform)

        if config.SETTINGS.main.import_cabling == "lldp":

            try:
                result = task.run(task=napalm_get, getters=["lldp_neighbors"])
            except:  # noqa: E722 # pylint: disable=bare-except
                LOGGER.debug("An exception occured while pulling lldp_data", exc_info=True)
                return Result(host=task.host, failed=True)

            if result[0].failed:
                return result

            neighbors = result[0].result.get("lldp_neighbors", {})
            return Result(host=task.host, result={"neighbors": neighbors})

        if config.SETTINGS.main.import_cabling == "cdp":

            try:
                result = task.run(task=netmiko_send_command, command_string="show cdp neighbors detail", use_genie=True)
            except NornirSubTaskError:
                LOGGER.debug("An exception occured while pulling CDP data")
                return Result(host=task.host, failed=True)

            if result[0].failed:
                return result

            results = convert_cisco_genie_cdp_neighbors_details(device_name=task.host.name, data=result[0].result)
            return Result(host=task.host, result=results.dict())

        LOGGER.warning("%s | Unexpected value for `import_cabling`, should be either LLDP or CDP", task.host.name)
        return Result(host=task.host, failed=True)

    @staticmethod
    def get_vlans(task: Task) -> Result:
        """Placeholder for get_vlans Get a list of vlans from the device.

        Args:
            task (Task): Nornir Task

        Returns:
            Result: Nornir Result object with a dict as a result containing the vlans
            The format of the result but must be similar to Vlans defined in network_importer.processors.get_vlans
        """
        LOGGER.warning("%s | Get Vlans not implemented in the default driver.", task.host.name)



"""GetConfig processor for the network_importer.

(c) 2020 Network To Code

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
  http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import os
import logging
from pathlib import Path

import hashlib

from nornir.core.inventory import Host
from nornir.core.task import AggregatedResult, MultiResult, Task

import network_importer.config as config
from network_importer.processors import BaseProcessor

LOGGER = logging.getLogger("network-importer")


class GetConfig(BaseProcessor):
    """GetConfig processor for the network_importer."""

    task_name = "get_config"
    config_extension = "txt"

    def __init__(self) -> None:
        """Initialize the processor and ensure some variables are properly initialized."""
        self.current_md5 = dict()
        self.previous_md5 = dict()
        self.config_filename = dict()
        self.config_dir = None
        self.existing_config_hostnames = None

    def task_started(self, task: Task) -> None:
        """Execute some house keeping item at the beginning at the execution.

        Before Update all the config file:
        - ensure that the configs directory exist
        - check what config files are already present

        Args:
            task (Task): Nornir Task
        """
        if not os.path.isdir(config.SETTINGS.main.configs_directory):
            os.mkdir(config.SETTINGS.main.configs_directory)
            LOGGER.debug("Configs directory created at %s", config.SETTINGS.main.configs_directory)

        self.config_dir = config.SETTINGS.main.configs_directory + "/configs"

        if not os.path.isdir(self.config_dir):
            os.mkdir(self.config_dir)
            LOGGER.debug("Configs directory created at %s", self.config_dir)

        # Save the hostnames associated with all existing configurations before we start the update process
        self.existing_config_hostnames = [
            f.split(f".{self.config_extension}")[0] for f in os.listdir(self.config_dir) if f.endswith(".txt")
        ]

    def task_completed(self, task: Task, result: AggregatedResult) -> None:
        """At the end, remove all configs files that have not been updated."""
        if len(self.existing_config_hostnames) > 0:
            LOGGER.info("Will delete %s config(s) that have not been updated", len(self.existing_config_hostnames))

            for hostname in self.existing_config_hostnames:
                os.remove(os.path.join(self.config_dir, f"{hostname}.{self.config_extension}"))

    def subtask_instance_started(self, task: Task, host: Host) -> None:
        """Before getting the new configuration, check if a configuration already exist and calculate its md5.

        Args:
            task (Task): Nornir Task
            host (Host): Nornir Host
        """
        if task.name != self.task_name:
            return

        self.config_filename[host.name] = f"{self.config_dir}/{task.host.name}.{self.config_extension}"

        if os.path.exists(self.config_filename[host.name]):
            current_config = Path(self.config_filename[host.name]).read_text()
            self.previous_md5[host.name] = hashlib.md5(current_config.encode("utf-8")).hexdigest()

    def subtask_instance_completed(self, task: Task, host: Host, result: MultiResult) -> None:
        """Verify the configuration returned and store it to disk.

        After collecting the configuration for each host.
        - Inspect the result and ensure the configuration is valid
        - If the configuration is valid, check the md5
        - if the MD5 is different, write the new configuration to disk

        Args:
            task (Task): Nornir Task
            host (Host): Nornir Host
            result (MultiResult): Nornir MultiResult
        """
        if task.name != self.task_name:
            return

        if result[0].failed:
            if result[0].exception:
                LOGGER.warning("%s | %s", task.host.name, result[0].exception)
            else:
                LOGGER.warning("%s | Something went wrong while trying to update the configuration ", task.host.name)
            host.status = "fail-other"
            return

        conf = result[0].result.get("config", None)

        if not conf:
            LOGGER.warning("%s | No configuration return ", task.host.name)
            host.status = "fail-other"
            return

        # Count the number of lines in the config file, if less than 10 report an error
        # mostlikely something went wrong while pulling the config
        if conf.count("\n") < 10:
            LOGGER.warning("%s | Less than 10 configuration lines returned", task.host.name)
            host.status = "fail-other"
            return

        if host.name in self.existing_config_hostnames:
            self.existing_config_hostnames.remove(host.name)

        # Save configuration to file and verify the new MD5
        with open(self.config_filename[host.name], "w") as config_:
            config_.write(conf)

        host.has_config = True

        self.current_md5[host.name] = hashlib.md5(conf.encode("utf-8")).hexdigest()
        # changed = False

        if host.name in self.previous_md5 and self.previous_md5[host.name] == self.current_md5[host.name]:
            LOGGER.info("%s | Latest config file already present ...", task.host.name)
        else:
            LOGGER.info("%s | Configuration file updated", task.host.name)
            # changed = True



"""BaseProcessor for the nornir.

(c) 2020 Network To Code

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
  http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import logging

from nornir.core.inventory import Host
from nornir.core.task import AggregatedResult, MultiResult, Task

logger = logging.getLogger("network-importer")

# pylint: disable=missing-function-docstring


class BaseProcessor:
    """Base Processor for nornir."""

    task_name = "'no task defined'"

    def task_started(self, task: Task) -> None:
        pass

    def task_completed(self, task: Task, result: AggregatedResult) -> None:
        pass

    def task_instance_started(self, task: Task, host: Host) -> None:
        pass

    def task_instance_completed(self, task: Task, host: Host, result: MultiResult) -> None:
        pass

    def subtask_instance_started(self, task: Task, host: Host) -> None:
        pass

    def subtask_instance_completed(self, task: Task, host: Host, result: MultiResult) -> None:
        pass



"""library of utilities.

(c) 2020 Network To Code

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
  http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import re
import logging
from urllib3 import connectionpool, poolmanager
import yaml

LOGGER = logging.getLogger("network-importer")  # pylint: disable=C0103


def patch_http_connection_pool(**constructor_kwargs):
    """This allows to override the default parameters of the HTTPConnectionPool constructor.

    For example, to increase the poolsize to fix problems
    with "HttpConnectionPool is full, discarding connection"
    call this function with maxsize=16 (or whatever size
    you want to give to the connection pool)

    Args:
      **constructor_kwargs:
    """

    class MyHTTPConnectionPool(connectionpool.HTTPConnectionPool):
        """Class to increase the size of the HTTP Connection pool."""

        def __init__(self, *args, **kwargs):
            """Initialize the HTTP Connection pool."""
            kwargs.update(constructor_kwargs)
            super().__init__(*args, **kwargs)

    poolmanager.pool_classes_by_scheme["http"] = MyHTTPConnectionPool


def sort_by_digits(if_name: str) -> tuple:
    """Extract all digits from a string and return them as tuple.

    Args:
      if_name (str): name of an interface

    Returns:
      tuple of all digits in the string
    """
    find_digit = re.compile(r"\D?(\d+)\D?")
    return tuple(map(int, find_digit.findall(if_name)))


def is_interface_physical(name):  # pylint: disable=R0911
    """Function evaluate if an interface is likely to be a physical interface.

    Args:
      name (str): name of the interface to evaluate

    Return:
      True, False or None
    """
    # Match most physical interface Cisco that contains Ethernet
    #  GigabitEthernet0/0/2
    #  GigabitEthernet0/0/2:3
    #  TenGigabitEthernet0/0/4
    cisco_physical_intf = r"^[a-zA-Z]+[Ethernet][0-9\/\:]+$"

    # Match Sub interfaces finishing with ".<number>"
    sub_intf = r".*\.[0-9]+$"

    # Regex for loopback and vlan interface
    loopback = r"^(L|l)(oopback|o)[0-9]+$"
    vlan = r"^(V|v)(lan)[0-9]+$"

    # Generic physical interface match
    #  mainly looking for <int>/<int> or <int>/<int>/<int> at the end
    generic_physical_intf = r"^[a-zA-Z\-]+[0-9]+\/[0-9\/\:]+$"

    # Match Juniper Interfaces
    jnpr_physical_intf = r"^[a-z]+\-[0-9\/\:]+$"

    if re.match(loopback, name):
        return False
    if re.match(vlan, name):
        return False
    if re.match(cisco_physical_intf, name):
        return True
    if re.match(sub_intf, name):
        return False
    if re.match(jnpr_physical_intf, name):
        return True
    if re.match(generic_physical_intf, name):
        return True

    return None


def is_interface_lag(name):
    """Function to evaluate if an interface is likely to be a lag.

    Args:
      name (str): name of the interface to evaluate

    Return:
      True, False or None
    """
    port_channel_intf = r"^port\-channel[0-9]+$"
    po_intf = r"^po[0-9]+$"
    ae_intf = r"^ae[0-9]+$"
    bundle_intf = r"^Bundle\-Ether[0-9]+$"

    if re.match(port_channel_intf, name.lower()):
        return True
    if re.match(ae_intf, name):
        return True
    if re.match(po_intf, name):
        return True
    if re.match(bundle_intf, name):
        return True

    return None


def is_mac_address(data):
    """Evaluate if a given string is a mac address.

    Args:
        data (str): string to evaluate

    Returns:
      bool: True if the string provided is a mac address, false otherwise
    """
    mac_address_chars = r"^[0-9a-fA-F\.\-\:]+$"
    hex_chars = r"[0-9a-fA-F]"

    if not isinstance(data, str):
        raise TypeError("data should be of type string")

    if not re.match(mac_address_chars, data):
        return False

    hex_data = re.findall(hex_chars, data)
    if len(hex_data) == 12:
        return True

    return False


def jinja_filter_toyaml_list(value) -> str:
    """Jinjafilter to return a list as a Nice Yaml.

    Args:
        value (list): value to convert

    Returns:
        Str formatted as Yaml
    """
    return yaml.dump(value, default_flow_style=None)


def jinja_filter_toyaml_dict(value) -> str:
    """Jinjafilter to return a dict as a Nice Yaml.

    Args:
        value (dict): value to convert

    Returns:
        Str formatted as Yaml
    """
    return yaml.dump(value, default_flow_style=False)


def expand_vlans_list(vlans: str) -> list:
    """Convert string of comma separated integer (vlan) into a list.

    Args:
        vlans (str)

    Returns:
        List: sorted list of vlans
    """
    raw_vlans_list = []
    clean_vlans_list = []

    vlans_csv = str(vlans).split(",")

    for vlan in vlans_csv:
        min_max = str(vlan).split("-")
        if len(min_max) == 1:
            raw_vlans_list.append(vlan)
        elif len(min_max) == 2:
            raw_vlans_list.extend(range(int(min_max[0]), int(min_max[1]) + 1))

    for vlan_ in raw_vlans_list:
        try:
            clean_vlans_list.append(int(vlan_))
        except ValueError as exc:
            LOGGER.debug("expand_vlans_list() Unable to convert %s as integer .. skipping (%s)", vlan_, exc)

    return sorted(clean_vlans_list)


def build_filter_params(filter_params, params):
    """Update params dict() with filter args in required format for pynetbox/pynautobot.

    Args:
      filter_params (list): split string from cli or config
      params (dict): object to hold params
    """
    for param_value in filter_params:
        if "=" not in param_value:
            continue
        key, value = param_value.split("=", 1)
        existing_value = params.get(key)
        if existing_value and isinstance(existing_value, list):
            params[key].append(value)
        elif existing_value and isinstance(existing_value, str):
            params[key] = [existing_value, value]
        else:
            params[key] = value



"""Settings definition for the network importer."""
# pylint: disable=invalid-name,redefined-outer-name

import os
import sys
from pathlib import Path
from typing import List, Dict, Optional, Union

import toml
from pydantic import BaseSettings, ValidationError
from typing_extensions import Literal

from network_importer.exceptions import ConfigLoadFatalError

SETTINGS = None

DEFAULT_DRIVERS_MAPPING = {
    "default": "network_importer.drivers.default",
    "cisco_nxos": "network_importer.drivers.cisco_default",
    "cisco_ios": "network_importer.drivers.cisco_default",
    "cisco_xr": "network_importer.drivers.cisco_default",
    "juniper_junos": "network_importer.drivers.juniper_junos",
    "arista_eos": "network_importer.drivers.arista_eos",
}

DEFAULT_BACKENDS = {
    "netbox": {
        "inventory": "NetBoxAPIInventory",
        "adapter": "network_importer.adapters.netbox_api.adapter.NetBoxAPIAdapter",
    },
    "nautobot": {
        "inventory": "NautobotAPIInventory",
        "adapter": "network_importer.adapters.nautobot_api.adapter.NautobotAPIAdapter",
    },
}

# pylint: disable=global-statement


class BatfishSettings(BaseSettings):
    """Settings definition for the Batfish section of the configuration."""

    address: str = "localhost"
    network_name: str = "network-importer"
    snapshot_name: str = "latest"
    port_v1: int = 9997
    port_v2: int = 9996
    use_ssl: bool = False
    api_key: Optional[str]

    class Config:
        """Additional parameters to automatically map environment variable to some settings."""

        fields = {
            "address": {"env": "BATFISH_ADDRESS"},
            "network_name": {"env": "BATFISH_NETWORK_NAME"},
            "snapshot_name": {"env": "BATFISH_SNAPSHOT_NAME"},
            "api_key": {"env": "BATFISH_API_KEY"},
            "port_v1": {"env": "BATFISH_PORT_V1"},
            "port_v2": {"env": "BATFISH_PORT_V2"},
            "use_ssl": {"env": "BATFISH_USE_SSL"},
        }


class NetworkSettings(BaseSettings):
    """Settings definition for the Network section of the configuration."""

    login: Optional[str]
    password: Optional[str]
    enable: bool = True

    netmiko_extras: Optional[dict]
    napalm_extras: Optional[dict]

    fqdns: List[str] = list()  # List of valid FQDN that can be found in the network

    class Config:
        """Additional parameters to automatically map environment variable to some settings."""

        fields = {
            "login": {"env": "NETWORK_DEVICE_LOGIN"},
            "password": {"env": "NETWORK_DEVICE_PWD"},
            "enable": {"env": "NETWORK_DEVICE_ENABLE"},
        }


class LogsSettings(BaseSettings):
    """Settings definition for the Log section of the configuration."""

    level: Literal["debug", "info", "warning"] = "info"
    # directory: str = "logs"
    performance_log: bool = False
    performance_log_directory: str = "performance_logs"
    # change_log: bool = True
    # change_log_format: Literal[
    #     "jsonlines", "text"
    # ] = "text"  # dict(type="string", enum=["jsonlines", "text"], default="text"),
    # change_log_filename: str = "changelog"


class MainSettings(BaseSettings):
    """Settings definition for the Main section of the configuration."""

    import_ips: bool = True
    import_prefixes: bool = False
    import_cabling: Union[bool, Literal["lldp", "cdp", "config", "no"]] = "lldp"
    excluded_platforms_cabling: List[str] = list()

    import_vlans: Union[bool, Literal["config", "cli", "no"]] = "config"
    import_intf_status: bool = False

    nbr_workers: int = 25

    configs_directory: str = "configs"

    backend: Optional[Literal["nautobot", "netbox"]]
    """Only Netbox and Nautobot backend are included by default, if you want to use another backend
    you must leave backend empty and define inventory.inventory_class and adapters.sot_class manually."""

    # NOT SUPPORTED CURRENTLY
    generate_hostvars: bool = False
    hostvars_directory: str = "host_vars"


class AdaptersSettings(BaseSettings):
    """Settings definition for the Adapters section of the configuration."""

    network_class: str = "network_importer.adapters.network_importer.adapter.NetworkImporterAdapter"
    network_settings: Optional[dict]

    sot_class: Optional[str]
    sot_settings: Optional[dict]


class DriversSettings(BaseSettings):
    """Settings definition for the Drivers section of the configuration."""

    mapping: Dict[str, str] = DEFAULT_DRIVERS_MAPPING


class InventorySettings(BaseSettings):
    """Settings definition for the Inventory section of the configuration.

    By default, the inventory will use the primary IP to reach out to the devices
    if the use_primary_ip flag is disabled, the inventory will try to use the hostname to the device
    """

    inventory_class: Optional[str]
    settings: Optional[dict]

    supported_platforms: List[str] = list()


class Settings(BaseSettings):
    """Main Settings Class for the project.

    The type of each setting is defined using Python annotations
    and is validated when a config file is loaded with Pydantic.
    """

    main: MainSettings = MainSettings()
    batfish: BatfishSettings = BatfishSettings()
    logs: LogsSettings = LogsSettings()
    network: NetworkSettings = NetworkSettings()
    adapters: AdaptersSettings = AdaptersSettings()
    drivers: DriversSettings = DriversSettings()
    inventory: InventorySettings = InventorySettings()


def _configure_backend(settings: Settings):
    """Configure the inventory and the SOT adapter for a given backend.

    The inventory and/or the adapter will be updated if:
      - a backend provided is valid
      - no value has been defined already

    Args:
        settings (Settings)

    Returns:
        Settings
    """
    if not settings.main.backend and (not settings.inventory.inventory_class or not settings.adapters.sot_class):
        raise ConfigLoadFatalError(
            "You must define a valid backend or assign inventory.inventory_class and adapters.sot_class manually."
        )

    if not settings.main.backend:
        return settings

    supported_backends = DEFAULT_BACKENDS.keys()
    if settings.main.backend not in supported_backends:
        raise ConfigLoadFatalError(f"backend value one of : {', '.join(supported_backends)}")

    if not settings.inventory.inventory_class:
        settings.inventory.inventory_class = DEFAULT_BACKENDS[settings.main.backend]["inventory"]

    if not settings.adapters.sot_class:
        settings.adapters.sot_class = DEFAULT_BACKENDS[settings.main.backend]["adapter"]

    return settings


def load(config_file_name="network_importer.toml", config_data=None):
    """Load configuration.

    Configuration is loaded from a file in network_importer.toml format that contains the settings,
    or from a dictionary of those settings passed in as "config_data"

    Args:
        config_file_name (str, optional): Name of the configuration file to load. Defaults to "network_importer.toml".
        config_data (dict, optional): dict to load as the config file instead of reading the file. Defaults to None.
    """
    global SETTINGS

    if config_data:
        SETTINGS = _configure_backend(Settings(**config_data))
        return

    if os.path.exists(config_file_name):
        config_string = Path(config_file_name).read_text()
        config_tmp = toml.loads(config_string)
        SETTINGS = _configure_backend(Settings(**config_tmp))
        return

    SETTINGS = Settings()


def load_and_exit(config_file_name="network_importer.toml", config_data=None):
    """Calls load, but wraps it in a try except block.

    This is done to handle a ValidationError which is raised when settings are specified but invalid.
    In such cases, a message is printed to the screen indicating the settings which don't pass validation.

    Args:
        config_file_name (str, optional): [description]. Defaults to "network_importer.toml".
        config_data (dict, optional): [description]. Defaults to None.
    """
    try:
        load(config_file_name=config_file_name, config_data=config_data)
    except ValidationError as err:
        print(f"Configuration not valid, found {len(err.errors())} error(s)")
        for error in err.errors():
            print(f"  {'/'.join(error['loc'])} | {error['msg']} ({error['type']})")
        sys.exit(1)
    except ConfigLoadFatalError as err:
        print("Configuration not valid")
        print(f"  {err}")
        sys.exit(1)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['network_importer',
 'network_importer.adapters',
 'network_importer.adapters.nautobot_api',
 'network_importer.adapters.netbox_api',
 'network_importer.adapters.network_importer',
 'network_importer.drivers',
 'network_importer.processors']

package_data = \
{'': ['*'], 'network_importer': ['templates/*']}

install_requires = \
['click>=7.1,<9.0',
 'diffsync>=1.2,<2.0',
 'genie>=21.4.0,<21.5.0',
 'netmiko>=3.3,<4.0',
 'nornir-napalm>=0.1.2,<0.2.0',
 'nornir-netmiko>=0.1.1,<0.2.0',
 'nornir-utils>=0.1.2,<0.2.0',
 'nornir>=3.0,<4.0',
 'ntc-templates>=2.0,<3.0',
 'pyats>=21.4.0,<21.5.0',
 'pybatfish==2021.2.26.830',
 'pydantic>=1.6,<2.0',
 'pynautobot>=1.0.2,<2.0.0',
 'pynetbox>=5.0,<6.0',
 'rich>=9.2,<10.0',
 'structlog>=20.1.0,<21.0.0',
 'termcolor>=1.1,<2.0',
 'toml>=0.10,<0.11']

entry_points = \
{'console_scripts': ['network-importer = network_importer.cli:main']}

setup_kwargs = {
    'name': 'network-importer',
    'version': '3.0.5',
    'description': 'Network Importer tool to import an existing network into a Database / Source Of Truth',
    'long_description': "# Network Importer\n\nThe network importer is a tool/library to analyze and/or synchronize an existing network with a Network Source of Truth (SOT), it's designed to be idempotent and by default it's only showing the difference between the running network and the remote SOT.\n\nThe main use cases for the network importer:\n\n- Import an existing network into a SOT (Nautobot or NetBox) as a first step to automate a brownfield network\n- Check the differences between the running network and the Source of Truth\n\n## Quick Start\n\n- [Getting Started](docs/getting_started.md)\n- [Configuration file](docs/configuration.md)\n- [Supported Features and Architecture](docs/architecture.md)\n- [Extensibility](docs/extensibility.md)\n- [Upgrade procedure](docs/upgrade.md)\n\n## Questions\n\nFor any questions or comments, please feel free to open an issue or swing by the [networktocode slack channel](https://networktocode.slack.com/). (channel #networktocode)\n\nSign up [here](http://slack.networktocode.com/)\n",
    'author': 'Network to Code, LLC',
    'author_email': 'opensource@networktocode.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/networktocode/network-importer',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)



"""DiffSync Models for the network importer.

(c) 2020 Network To Code

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
  http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from typing import List, Optional

from diffsync import DiffSyncModel


class Site(DiffSyncModel):
    """Site Model based on DiffSyncModel.

    A site must have a unique name and can be composed of Vlans and Prefixes.
    """

    _modelname = "site"
    _identifiers = ("name",)
    _children = {"vlan": "vlans", "prefix": "prefixes"}

    name: str
    prefixes: List = list()
    vlans: List[str] = list()


class Device(DiffSyncModel):
    """Device Model based on DiffSyncModel.

    A device must have a unique name and can be part of a site.
    """

    _modelname = "device"
    _identifiers = ("name",)
    _attributes = ("site_name",)
    _children = {"interface": "interfaces"}

    name: str
    site_name: Optional[str]
    interfaces: List = list()

    platform: Optional[str]
    model: Optional[str]
    role: Optional[str]
    vendor: Optional[str]


class Interface(DiffSyncModel):  # pylint: disable=too-many-instance-attributes
    """Interface Model based on DiffSyncModel.

    An interface must be attached to a device and the name must be unique per device.
    """

    _modelname = "interface"
    _identifiers = ("device_name", "name")
    _shortname = ("name",)
    _attributes = (
        "description",
        # "mtu",
        "is_virtual",
        "is_lag",
        "is_lag_member",
        "parent",
        "mode",
        "switchport_mode",
        "allowed_vlans",
        "access_vlan",
    )
    _children = {"ip_address": "ips"}

    name: str
    device_name: str

    description: Optional[str]
    mtu: Optional[int]
    speed: Optional[int]
    mode: Optional[str]  # TRUNK, ACCESS, L3, NONE
    switchport_mode: Optional[str] = "NONE"
    active: Optional[bool]
    is_virtual: Optional[bool]
    is_lag: Optional[bool]
    is_lag_member: Optional[bool]
    parent: Optional[str]

    lag_members: List[str] = list()
    allowed_vlans: List[str] = list()
    access_vlan: Optional[str]

    ips: List[str] = list()


class IPAddress(DiffSyncModel):
    """IPAddress Model based on DiffSyncModel.

    An IP address must be unique and can be associated with an interface.
    """

    _modelname = "ip_address"
    _identifiers = ("device_name", "interface_name", "address")

    device_name: str
    interface_name: str
    address: str


class Prefix(DiffSyncModel):
    """Prefix Model based on DiffSyncModel.

    An Prefix must be associated with a Site and must be unique within a site.
    """

    _modelname = "prefix"
    _identifiers = ("site_name", "prefix")
    _attributes = ("vlan",)

    prefix: str
    site_name: Optional[str]
    vlan: Optional[str]


class Cable(DiffSyncModel):
    """Cable Model based on DiffSyncModel."""

    _modelname = "cable"
    _identifiers = (
        "device_a_name",
        "interface_a_name",
        "device_z_name",
        "interface_z_name",
    )

    device_a_name: str
    interface_a_name: str
    device_z_name: str
    interface_z_name: str

    source: Optional[str]
    is_valid: bool = True
    error: Optional[str]

    def __init__(self, *args, **kwargs):
        """Ensure the cable is unique by ordering the devices alphabetically."""
        if "device_a_name" not in kwargs or "device_z_name" not in kwargs:
            raise ValueError("device_a_name and device_z_name are mandatory")
        if not kwargs["device_a_name"] or not kwargs["device_z_name"]:
            raise ValueError("device_a_name and device_z_name are mandatory and must not be None")

        keys_to_copy = ["device_a_name", "interface_a_name", "device_z_name", "interface_z_name"]
        ids = {key: kwargs[key] for key in keys_to_copy}

        devices = [kwargs["device_a_name"], kwargs["device_z_name"]]
        if sorted(devices) != devices:
            ids["device_a_name"] = kwargs["device_z_name"]
            ids["interface_a_name"] = kwargs["interface_z_name"]
            ids["device_z_name"] = kwargs["device_a_name"]
            ids["interface_z_name"] = kwargs["interface_a_name"]

        for key in keys_to_copy:
            del kwargs[key]

        super().__init__(*args, **ids, **kwargs)

    def get_device_intf(self, side):
        """Get the device name and the interface name for a given side.

        Args:
            side (str): site to query, must be either a or z

        Raises:
            ValueError: when the side is not either a or z

        Returns:
            (device_name (str), interface_name (str))
        """
        if side.lower() == "a":
            return self.device_a_name, self.interface_a_name

        if side.lower() == "z":
            return self.device_z_name, self.interface_z_name

        raise ValueError("side must be either 'a' or 'z'")


class Vlan(DiffSyncModel):
    """Vlan Model based on DiffSyncModel.

    An Vlan must be associated with a Site and the vlan_id msut be unique within a site.
    """

    _modelname = "vlan"
    _identifiers = ("site_name", "vid")
    _attributes = ("name", "associated_devices")

    vid: int
    site_name: str
    name: Optional[str]

    associated_devices: List[str] = list()

    def add_device(self, device_name):
        """Add a device to the list of associated devices.

        Args:
            device_name (str): name of a device to associate with this VLAN
        """
        if device_name not in self.associated_devices:
            self.associated_devices.append(device_name)
            self.associated_devices = sorted(self.associated_devices)



"""GetNeighbors processor for the network_importer.

(c) 2020 Network To Code

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
  http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from collections import defaultdict
import logging
import re

from typing import Dict, List

from nornir.core.inventory import Host
from nornir.core.task import MultiResult, Task
from pydantic import BaseModel  # pylint: disable=no-name-in-module

import network_importer.config as config
from network_importer.processors import BaseProcessor
from network_importer.utils import is_mac_address

LOGGER = logging.getLogger("network-importer")

# Possible Junos port names xe-0/0/1.0, xe-0/0/3:0, ge, et, em sxe, fte, me, fc, xle
# Match the incorrectly capitalized interface names
JUNOS_INTERFACE_PATTERN = re.compile(r"^(Xe|Ge|Et|Em|Sxe|Fte|Me|Fc|Xle)-\d+/\d+/\d+[.:]*\d*$")


# -----------------------------------------------------------------
# Inventory Filter functions
# -----------------------------------------------------------------
def hosts_for_cabling(host):
    """Inventory Filter for Nornir.

    Return True or False if a device is eligible for cabling.
    It's using config.SETTINGS.main.excluded_platforms_cabling to determine if a host is eligible.

    Args:
      host(Host): Nornir Host

    Returns:
        bool: True if the device is eligible for cabling, False otherwise.
    """
    if host.platform in config.SETTINGS.main.excluded_platforms_cabling:
        return False

    return True


# -----------------------------------------------------------------
# Expected Returned Data
# -----------------------------------------------------------------
class Neighbor(BaseModel):
    """Dataclass model to store one Neighbor returned by get_neighbors."""

    hostname: str
    port: str


class Neighbors(BaseModel):
    """Dataclass model to store all Neighbors returned by get_neighbors."""

    neighbors: Dict[str, List[Neighbor]] = defaultdict(list)


# -----------------------------------------------------------------
# Processor
# -----------------------------------------------------------------
class GetNeighbors(BaseProcessor):
    """GetNeighbors processor for the network_importer."""

    task_name = "get_neighbors"

    def subtask_instance_started(self, task: Task, host: Host) -> None:
        """Before getting the new configuration, check if a configuration already exist and calculate its md5.

        Args:
            task (Task): Nornir Task
            host (Host): Nornir Host
        """
        if task.name != self.task_name:
            return

    def subtask_instance_completed(self, task: Task, host: Host, result: MultiResult) -> None:
        """For each host, check if the results returned by the task is valid and cleanup results as needed.

        Args:
            task (Task): Nornir Task
            host (Host): Nornir Host
            result (MultiResult): Nornir Results
        """
        if task.name != self.task_name:
            return

        if result[0].failed:
            LOGGER.warning("%s | Something went wrong while trying to pull the neighbor information", host.name)
            host.status = "fail-other"
            return

        if not isinstance(result[0].result, dict) or "neighbors" not in result[0].result:
            LOGGER.warning("%s | No neighbor information returned", host.name)
            result[0].failed = True
            return

        interfaces = list(result[0].result["neighbors"].keys())
        for interface in interfaces:

            neighbors = result[0].result["neighbors"][interface]

            if len(neighbors) > 1:
                LOGGER.warning("%s | More than 1 neighbor found on interface %s, SKIPPING", host.name, interface)
                del result[0].result["neighbors"][interface]
                continue

            if is_mac_address(neighbors[0]["hostname"]):
                del result[0].result["neighbors"][interface]
                continue

            # Clean up hostname to remove full FQDN
            result[0].result["neighbors"][interface][0]["hostname"] = self.clean_neighbor_name(neighbors[0]["hostname"])

            # Clean up the portname if genie incorrectly capitalized it
            result[0].result["neighbors"][interface][0]["port"] = self.clean_neighbor_port_name(neighbors[0]["port"])

    @classmethod
    def clean_neighbor_name(cls, neighbor_name):
        """Cleanup the name of a neighbor by removing all known FQDNs.

        Args:
            neighbor_name ([str]): name of a neighbor returned by cdp or lldp

        Returns:
            str: clean neighboar name
        """
        # Remove all FQDN from the hostname to match what is in the SOT
        config.SETTINGS.network.fqdns.sort(key=len, reverse=True)
        for fqdn in config.SETTINGS.network.fqdns:
            if fqdn in neighbor_name:
                return neighbor_name.replace(f".{fqdn}", "")

        return neighbor_name

    @classmethod
    def clean_neighbor_port_name(cls, port_name):
        """Work around for https://github.com/CiscoTestAutomation/genieparser/issues/287."""
        if JUNOS_INTERFACE_PATTERN.match(port_name):
            port_name = port_name[0].lower() + port_name[1:]
            return port_name

        return port_name



"""Extension of the base Models for the NetboxAPIAdapter."""
from typing import Optional
import logging

import pynetbox
from diffsync.exceptions import ObjectNotFound
from diffsync import DiffSync, DiffSyncModel  # pylint: disable=unused-import

import network_importer.config as config  # pylint: disable=import-error
from network_importer.adapters.netbox_api.exceptions import NetboxObjectNotValid
from network_importer.models import (  # pylint: disable=import-error
    Site,
    Device,
    Interface,
    IPAddress,
    Cable,
    Prefix,
    Vlan,
)

LOGGER = logging.getLogger("network-importer")


class NetboxSite(Site):
    """Extension of the Site model."""

    remote_id: Optional[int]


class NetboxDevice(Device):
    """Extension of the Device model."""

    remote_id: Optional[int]
    primary_ip: Optional[str]

    device_tag_id: Optional[int]

    def get_device_tag_id(self):
        """Get the Netbox id of the tag for this device.

        If the ID is already present locally return it
        If not try to retrieve it from netbox or create it in Netbox if needed

        Returns:
            device_tag_id (int)
        """
        if self.device_tag_id:
            return self.device_tag_id

        tag = self.diffsync.netbox.extras.tags.get(name=f"device={self.name}")

        if not tag:
            tag = self.diffsync.netbox.extras.tags.create(name=f"device={self.name}", slug=f"device__{self.name}")

        self.device_tag_id = tag.id
        return self.device_tag_id


class NetboxInterface(Interface):
    """Extension of the Interface model."""

    remote_id: Optional[int]
    connected_endpoint_type: Optional[str]

    def translate_attrs_for_netbox(self, attrs):  # pylint: disable=too-many-branches
        """Translate interface attributes into Netbox format.

        Args:
            params (dict): Dictionnary of attributes of the object to translate

        Returns:
            dict: Netbox parameters
        """

        def convert_vlan_to_nid(vlan_uid):
            if not vlan_uid:
                return None
            try:
                vlan = self.diffsync.get(self.diffsync.vlan, identifier=vlan_uid)
            except ObjectNotFound:
                return None

            return vlan.remote_id

        def convert_vlan_list_to_nids(vlan_uids):
            resp = []
            for uid in vlan_uids:
                nid = convert_vlan_to_nid(uid)
                if nid:
                    resp.append(nid)
            return resp

        # Reconstruct all attrs of the object in its expected state
        complete_attrs = self.get_attrs()
        complete_attrs.update(attrs)
        attrs = complete_attrs

        nb_params = {}

        # Identify the id of the device this interface is attached to
        device = self.diffsync.get(self.diffsync.device, identifier=self.device_name)

        if not device.remote_id:
            raise NetboxObjectNotValid(f"device {self.device_name}, is missing a remote_id")

        nb_params["device"] = device.remote_id
        nb_params["name"] = self.name

        if "is_lag" in attrs and attrs["is_lag"]:
            nb_params["type"] = "lag"
        elif "is_virtual" in attrs and attrs["is_virtual"]:
            nb_params["type"] = "virtual"
        else:
            nb_params["type"] = "other"

        if "mtu" in attrs:
            nb_params["mtu"] = attrs["mtu"]

        if "description" in attrs:
            nb_params["description"] = attrs["description"] or ""

        if "switchport_mode" in attrs and attrs["switchport_mode"] == "ACCESS":
            nb_params["mode"] = "access"
        elif "switchport_mode" in attrs and attrs["switchport_mode"] == "TRUNK":
            nb_params["mode"] = "tagged"

        # if is None:
        #     intf_properties["enabled"] = intf.active

        if config.SETTINGS.main.import_vlans not in [False, "no"]:
            if "mode" in attrs and attrs["mode"] in ["TRUNK", "ACCESS"] and "access_vlan" in attrs:
                nb_params["untagged_vlan"] = convert_vlan_to_nid(attrs["access_vlan"])

            if (
                "mode" in attrs
                and attrs["mode"] in ["TRUNK", "L3_SUB_VLAN"]
                and "allowed_vlans" in attrs
                and attrs["allowed_vlans"]
            ):
                nb_params["tagged_vlans"] = convert_vlan_list_to_nids(attrs["allowed_vlans"])
            elif (
                "mode" in attrs
                and attrs["mode"] in ["TRUNK", "L3_SUB_VLAN"]
                and ("allowed_vlans" not in attrs or not attrs["allowed_vlans"])
            ):
                nb_params["tagged_vlans"] = []

        if "is_lag_member" in attrs and attrs["is_lag_member"] and "parent" in attrs:
            try:
                parent_interface = self.diffsync.get(self.diffsync.interface, identifier=attrs["parent"])
                if parent_interface and parent_interface.remote_id:
                    nb_params["lag"] = parent_interface.remote_id
            except ObjectNotFound:
                LOGGER.warning("No Parent interface found for lag member %s %s", self.device_name, self.name)
                nb_params["lag"] = None

        elif "is_lag_member" in attrs and not attrs["is_lag_member"]:
            nb_params["lag"] = None

        return nb_params

    @classmethod
    def create(cls, diffsync: "DiffSync", ids: dict, attrs: dict) -> Optional["DiffSyncModel"]:
        """Create an interface object in Netbox.

        Args:
            diffsync: The master data store for other DiffSyncModel instances that we might need to reference
            ids: Dictionary of unique-identifiers needed to create the new object
            attrs: Dictionary of additional attributes to set on the new object

        Returns:
            NetboxInterface: DiffSync object newly created
        """
        item = super().create(ids=ids, diffsync=diffsync, attrs=attrs)

        try:
            nb_params = item.translate_attrs_for_netbox(attrs)
            intf = diffsync.netbox.dcim.interfaces.create(**nb_params)
            LOGGER.info("Created interface %s (%s) in NetBox", intf.name, intf.id)
        except pynetbox.core.query.RequestError as exc:
            LOGGER.warning(
                "Unable to create interface %s on %s in %s (%s)",
                ids["name"],
                ids["device_name"],
                diffsync.name,
                exc.error,
            )
            return item
        except Exception as exc:  # pylint: disable=broad-except
            LOGGER.warning(
                "Unable to create interface %s on %s in %s (%s)",
                ids["name"],
                ids["device_name"],
                diffsync.name,
                str(exc),
            )
            return item

        item.remote_id = intf.id
        return item

    def update(self, attrs: dict) -> Optional["DiffSyncModel"]:
        """Update an interface object in Netbox.

        Args:
            attrs: Dictionary of attributes to update on the object

        Returns:
            DiffSyncModel: this instance, if all data was successfully updated.
            None: if data updates failed in such a way that child objects of this model should not be modified.

        Raises:
            ObjectNotUpdated: if an error occurred.
        """
        current_attrs = self.get_attrs()

        if attrs == current_attrs:
            return self

        current_attrs.update(attrs)
        nb_params = self.translate_attrs_for_netbox(current_attrs)
        LOGGER.debug("Update interface : %s", nb_params)
        try:
            intf = self.diffsync.netbox.dcim.interfaces.get(self.remote_id)
            intf.update(data=nb_params)
            LOGGER.info("Updated Interface %s %s (%s) in NetBox", self.device_name, self.name, self.remote_id)
        except pynetbox.core.query.RequestError as exc:
            LOGGER.warning(
                "Unable to update interface %s on %s in %s (%s)",
                self.name,
                self.device_name,
                self.diffsync.name,
                exc.error,
            )
            return None

        return super().update(attrs)

    def delete(self) -> Optional["DiffSyncModel"]:
        """Delete an interface object in Netbox.

        Returns:
            NetboxInterface: DiffSync object
        """
        # Check if the interface has some Ips, check if it is the management interface
        if self.ips:
            try:
                dev = self.diffsync.get(self.diffsync.device, identifier=self.device_name)
                if dev.primary_ip and dev.primary_ip in self.ips:
                    LOGGER.warning(
                        "Unable to delete interface %s on %s, because it's currently the management interface",
                        self.name,
                        dev.name,
                    )
                    return self
            except ObjectNotFound:
                LOGGER.debug(
                    "Unable to delete interface %s on %s, because device is not present. See issue #188",
                    self.name,
                    self.device_name,
                )
                return None
        try:
            intf = self.diffsync.netbox.dcim.interfaces.get(self.remote_id)
            intf.delete()
        except pynetbox.core.query.RequestError as exc:
            LOGGER.warning(
                "Unable to delete Interface %s on %s in %s (%s)",
                self.name,
                self.device_name,
                self.diffsync.name,
                exc.error,
            )

        super().delete()
        return self


class NetboxIPAddress(IPAddress):
    """Extension of the IPAddress model."""

    remote_id: Optional[int]

    def translate_attrs_for_netbox(self, attrs=None):  # pylint: disable=unused-argument
        """Translate IP address attributes into NetBox format.

        Args:
            attrs (dict): Dictionnary of attributes of the object to translate

        Returns:
            dict: Netbox parameters
        """
        nb_params = {"address": self.address}

        try:
            interface = self.diffsync.get(
                self.diffsync.interface,
                identifier=dict(device_name=self.device_name, name=self.interface_name),
            )
            nb_params["assigned_object_type"] = "dcim.interface"
            nb_params["assigned_object_id"] = interface.remote_id
        except ObjectNotFound:
            pass

        return nb_params

    @classmethod
    def create_from_pynetbox(cls, diffsync: "DiffSync", obj, device_name):  # pylint: disable=unused-argument
        """Create a new NetboxIPAddress object from a pynetbox ip_address object.

        Args:
            diffsync (DiffSync): Netbox API Adapter
            obj (pynetbox.models.ipam.IpAddresses): IPAddress object returned by Pynetbox
            device_name (str): name of the device associated with this ip address
        Returns:
            NetboxIPAddress: DiffSync object
        """
        item = cls(
            address=obj.address, device_name=device_name, interface_name=obj.assigned_object.name, remote_id=obj.id
        )

        item = diffsync.apply_model_flag(item, obj)
        return item

    @classmethod
    def create(cls, diffsync: "DiffSync", ids: dict, attrs: dict) -> Optional["DiffSyncModel"]:
        """Create an IP address in Netbox, if the name of a valid interface is provided the interface will be assigned to the interface.

        Returns:
            NetboxIPAddress: DiffSync object
        """
        try:
            item = super().create(ids=ids, diffsync=diffsync, attrs=attrs)
            nb_params = item.translate_attrs_for_netbox(attrs)
            ip_address = diffsync.netbox.ipam.ip_addresses.create(**nb_params)
        except pynetbox.core.query.RequestError as exc:
            LOGGER.warning("Unable to create the ip address %s in %s (%s)", ids["address"], diffsync.name, exc.error)
            return None

        LOGGER.info("Created IP %s (%s) in NetBox", ip_address.address, ip_address.id)
        item.remote_id = ip_address.id

        return item

    def delete(self) -> Optional["DiffSyncModel"]:
        """Delete an IP address in NetBox.

        Returns:
            NetboxIPAddress: DiffSync object
        """
        if self.device_name:
            try:
                dev = self.diffsync.get(self.diffsync.device, identifier=self.device_name)
                if dev.primary_ip == self.address:
                    LOGGER.warning(
                        "Unable to delete IP Address %s on %s, because it's currently the management IP address",
                        self.address,
                        self.device_name,
                    )
                    return None
            except ObjectNotFound:
                LOGGER.debug(
                    "Unable to delete IP Address %s on %s, because device is not present. See issue #188",
                    self.address,
                    self.device_name,
                )
                return None
        try:
            ipaddr = self.diffsync.netbox.ipam.ip_addresses.get(self.remote_id)
            ipaddr.delete()
        except pynetbox.core.query.RequestError as exc:
            LOGGER.warning(
                "Unable to delete IP Address %s on %s in %s (%s)",
                self.address,
                self.device_name,
                self.diffsync.name,
                exc.error,
            )
            return None

        super().delete()
        return self


class NetboxIPAddressPre29(NetboxIPAddress):
    """Extension of the IPAddress model specific for version of NetBox prior to 2.9.

    The method to attach an ip address to an interface changed in 2.9, this class implement the old method
    """

    def translate_attrs_for_netbox(self, attrs=None):
        """Translate IP address attributes into NetBox 2.8 format.

        Args:
            attrs (dict, Optional): Dictionnary of attributes of the object to translate

        Returns:
            dict: Netbox parameters
        """
        nb_params = {"address": self.address}

        try:
            interface = self.diffsync.get(
                self.diffsync.interface,
                identifier=dict(device_name=self.device_name, name=self.interface_name),
            )
            nb_params["interface"] = interface.remote_id
        except ObjectNotFound:
            pass

        return nb_params

    @classmethod
    def create_from_pynetbox(cls, diffsync: "DiffSync", obj, device_name):
        """Create a new NetboxIPAddress object from a pynetbox ip_address object specific for version of NetBox prior to 2.9.

        Args:
            diffsync (DiffSync): Netbox API Adapter
            obj (pynetbox.models.ipam.IpAddresses): IPAddress object returned by Pynetbox
            device_name (str): name of the device associated with this ip address
        Returns:
            NetboxIPAddress: DiffSync object
        """
        item = cls(address=obj.address, device_name=device_name, interface_name=obj.interface.name, remote_id=obj.id)
        item = diffsync.apply_model_flag(item, obj)

        return item


class NetboxPrefix(Prefix):
    """Extension of the Prefix model."""

    remote_id: Optional[int]

    def translate_attrs_for_netbox(self, attrs):
        """Translate prefix attributes into Netbox format.

        Args:
            attrs (dict): Dictionnary of attributes of the object to translate

        Returns:
            dict: Netbox parameters
        """
        nb_params = {"prefix": self.prefix, "status": "active"}

        site = self.diffsync.get(self.diffsync.site, identifier=self.site_name)
        nb_params["site"] = site.remote_id

        if "vlan" in attrs and attrs["vlan"]:
            try:
                vlan = self.diffsync.get(self.diffsync.vlan, identifier=attrs["vlan"])
                if vlan.remote_id:
                    nb_params["vlan"] = vlan.remote_id
            except ObjectNotFound:
                pass

        return nb_params

    @classmethod
    def create(cls, diffsync: "DiffSync", ids: dict, attrs: dict) -> Optional["DiffSyncModel"]:
        """Create a Prefix in NetBox.

        Returns:
            NetboxPrefix: DiffSync object
        """
        item = super().create(ids=ids, diffsync=diffsync, attrs=attrs)
        nb_params = item.translate_attrs_for_netbox(attrs)

        try:
            prefix = diffsync.netbox.ipam.prefixes.create(**nb_params)
            LOGGER.info("Created Prefix %s (%s) in NetBox", prefix.prefix, prefix.id)
        except pynetbox.core.query.RequestError as exc:
            LOGGER.warning("Unable to create Prefix %s in %s (%s)", ids["prefix"], diffsync.name, exc.error)
            return None

        item.remote_id = prefix.id

        return item

    def update(self, attrs: dict) -> Optional["DiffSyncModel"]:
        """Update a Prefix object in Netbox.

        Args:
            attrs: Dictionary of attributes to update on the object

        Returns:
            DiffSyncModel: this instance, if all data was successfully updated.
            None: if data updates failed in such a way that child objects of this model should not be modified.

        Raises:
            ObjectNotUpdated: if an error occurred.
        """
        current_attrs = self.get_attrs()

        if attrs == current_attrs:
            return self

        nb_params = self.translate_attrs_for_netbox(attrs)

        try:
            prefix = self.diffsync.netbox.ipam.prefixes.get(self.remote_id)
            prefix.update(data=nb_params)
            LOGGER.info("Updated Prefix %s (%s) in NetBox", self.prefix, self.remote_id)
        except pynetbox.core.query.RequestError as exc:
            LOGGER.warning(
                "Unable to update perfix %s in %s (%s)",
                self.prefix,
                self.diffsync.name,
                exc.error,
            )
            return None

        return super().update(attrs)


class NetboxVlan(Vlan):
    """Extension of the Vlan model."""

    remote_id: Optional[int]
    tag_prefix: str = "device="

    def translate_attrs_for_netbox(self, attrs):
        """Translate vlan attributes into Netbox format.

        Args:
            attrs (dict): Dictionnary of attributes of the object to translate

        Returns:
            dict: Netbox parameters
        """
        nb_params = {"vid": self.vid}

        if "name" in attrs and attrs["name"]:
            nb_params["name"] = attrs["name"]
        elif not self.name:
            nb_params["name"] = f"vlan-{self.vid}"

        site = self.diffsync.get(self.diffsync.site, identifier=self.site_name)
        nb_params["site"] = site.remote_id

        if "associated_devices" in attrs:
            nb_params["tags"] = []
            for device_name in attrs["associated_devices"]:
                try:
                    device = self.diffsync.get(self.diffsync.device, identifier=device_name)
                except ObjectNotFound:
                    LOGGER.error(
                        "Found an associated device on Vlan %s that doesn't exist (%s)",
                        self.get_unique_id(),
                        device_name,
                    )
                    continue

                tag_id = device.get_device_tag_id()
                nb_params["tags"].append(tag_id)

        return nb_params

    @classmethod
    def create_from_pynetbox(cls, diffsync: "DiffSync", obj, site_name):
        """Create a new NetboxVlan object from a pynetbox vlan object.

        Args:
            diffsync (DiffSync): Netbox API Adapter
            obj (pynetbox.models.ipam.Vlans): Vlan object returned by Pynetbox
            site_name (str): name of the site associated with this vlan

        Returns:
            NetboxVlan: DiffSync object
        """
        item = cls(vid=obj.vid, site_name=site_name, name=obj.name, remote_id=obj.id)

        # Check the existing tags to learn which device is already associated with this vlan
        # Exclude all devices that are not part of the inventory
        for tag in obj.tags:
            if item.tag_prefix not in tag["name"]:
                continue

            device_name = tag["name"].split(item.tag_prefix)[1]
            try:
                device = diffsync.get(diffsync.device, identifier=device_name)
            except ObjectNotFound:
                device = None
            if device:
                item.add_device(device_name)
                device.device_tag_id = tag["id"]

        item = diffsync.apply_model_flag(item, obj)

        return item

    @classmethod
    def create(cls, diffsync: "DiffSync", ids: dict, attrs: dict) -> Optional["DiffSyncModel"]:
        """Create new Vlan in NetBox.

        Returns:
            NetboxVlan: DiffSync object
        """
        try:
            item = super().create(ids=ids, diffsync=diffsync, attrs=attrs)
            nb_params = item.translate_attrs_for_netbox(attrs)
            vlan = diffsync.netbox.ipam.vlans.create(**nb_params)
            item.remote_id = vlan.id
            LOGGER.info("Created Vlan %s in %s (%s)", item.get_unique_id(), diffsync.name, vlan.id)
        except pynetbox.core.query.RequestError as exc:
            LOGGER.warning("Unable to create Vlan %s in %s (%s)", ids, diffsync.name, exc.error)
            return None

        return item

    def update_clean_tags(self, nb_params, obj):
        """Update list of vlan tags with additinal tags that already exists on the object in netbox.

        Args:
            nb_params (dict): dict of parameters in netbox format
            obj (pynetbox): Vlan object from pynetbox
        """
        # Before updating the remote vlan we need to check the existing list of tags
        # to ensure that we won't delete an existing tags
        if "tags" in nb_params and nb_params["tags"] and obj.tags:
            for tag in obj.tags:
                if self.tag_prefix not in tag["name"]:
                    nb_params["tags"].append(tag["id"])
                else:
                    dev_name = tag["name"].split(self.tag_prefix)[1]
                    try:
                        res = self.diffsync.get(self.diffsync.device, identifier=dev_name)
                    except ObjectNotFound:
                        res = None
                    if not res:
                        nb_params["tags"].append(tag["id"])

        return nb_params

    def update(self, attrs: dict) -> Optional["DiffSyncModel"]:
        """Update new Vlan in NetBox.

        Returns:
            NetboxVlan: DiffSync object
        """
        nb_params = self.translate_attrs_for_netbox(attrs)

        try:
            vlan = self.diffsync.netbox.ipam.vlans.get(self.remote_id)
            clean_params = self.update_clean_tags(nb_params=nb_params, obj=vlan)
            vlan.update(data=clean_params)
            LOGGER.info("Updated Vlan %s (%s) in NetBox", self.get_unique_id(), self.remote_id)
        except pynetbox.core.query.RequestError as exc:
            LOGGER.warning("Unable to update Vlan %s in %s (%s)", self.get_unique_id(), self.diffsync.name, exc.error)
            return None

        return super().update(attrs)


class NetboxVlanPre29(NetboxVlan):
    """Extension of the Vlan model."""

    remote_id: Optional[int]
    tag_prefix: str = "device="

    def translate_attrs_for_netbox(self, attrs):
        """Translate vlan parameters into Netbox format.

        Args:
            params (dict): Dictionnary of attributes/parameters of the object to translate

        Returns:
            dict: Netbox parameters
        """
        nb_params = {"vid": self.vid}

        if "name" in attrs and attrs["name"]:
            nb_params["name"] = attrs["name"]
        elif not self.name:
            nb_params["name"] = f"vlan-{self.vid}"

        site = self.diffsync.get(self.diffsync.site, identifier=self.site_name)
        nb_params["site"] = site.remote_id

        if "associated_devices" in attrs:
            nb_params["tags"] = [f"device={device}" for device in attrs["associated_devices"]]

        return nb_params

    @classmethod
    def create_from_pynetbox(cls, diffsync: "DiffSync", obj, site_name):
        """Create a new NetboxVlan object from a pynetbox vlan object for version of NetBox prior to 2.9.

        Args:
            diffsync (DiffSync): Netbox API Adapter
            obj (pynetbox.models.ipam.Vlans): Vlan object returned by Pynetbox
            site_name (str): name of the site associated with this vlan

        Returns:
            NetboxVlan: DiffSync object
        """
        item = cls(vid=obj.vid, site_name=site_name, name=obj.name, remote_id=obj.id)

        # Check the existing tags to learn which device is already associated with this vlan
        # Exclude all vlans that are not part of the inventory
        if obj.tags:
            all_devices = [tag.split(item.tag_prefix)[1] for tag in obj.tags if item.tag_prefix in tag]
            for device in all_devices:
                try:
                    diffsync.get(diffsync.device, identifier=device)
                    item.add_device(device)
                except ObjectNotFound:
                    pass

        item = diffsync.apply_model_flag(item, obj)

        return item

    def update_clean_tags(self, nb_params, obj):
        """Update list of vlan tags with additional tags that already exists on the object in netbox.

        Args:
            nb_params (dict): dict of parameters in netbox format
            obj (pynetbox): Vlan object from pynetbox
        """
        # Before updating the remote vlan we need to check the existing list of tags
        # to ensure that we won't delete an existing tags
        if "tags" in nb_params and nb_params["tags"] and obj.tags:
            for tag in obj.tags:
                if self.tag_prefix not in tag:
                    nb_params["tags"].append(tag)
                else:
                    dev_name = tag.split(self.tag_prefix)[1]
                    try:
                        self.diffsync.get(self.diffsync.device, identifier=dev_name)
                    except ObjectNotFound:
                        nb_params["tags"].append(tag)

        return nb_params


class NetboxCable(Cable):
    """Extension of the Cable model."""

    remote_id: Optional[int]
    termination_a_id: Optional[int]
    termination_z_id: Optional[int]

    @classmethod
    def create(cls, diffsync: "DiffSync", ids: dict, attrs: dict) -> Optional["DiffSyncModel"]:
        """Create a Cable in NetBox.

        Returns:
            NetboxCable: DiffSync object
        """
        item = super().create(ids=ids, diffsync=diffsync, attrs=attrs)

        try:
            interface_a = diffsync.get(
                diffsync.interface, identifier=dict(device_name=ids["device_a_name"], name=ids["interface_a_name"])
            )
        except ObjectNotFound:
            interface_a = diffsync.get_intf_from_netbox(
                device_name=ids["device_a_name"], intf_name=ids["interface_a_name"]
            )
            if not interface_a:
                LOGGER.info(
                    "Unable to create Cable %s in %s, unable to find the interface %s %s",
                    item.get_unique_id(),
                    diffsync.name,
                    ids["device_a_name"],
                    ids["interface_a_name"],
                )
                return item

        try:
            interface_z = diffsync.get(
                diffsync.interface, identifier=dict(device_name=ids["device_z_name"], name=ids["interface_z_name"])
            )
        except ObjectNotFound:
            interface_z = diffsync.get_intf_from_netbox(
                device_name=ids["device_z_name"], intf_name=ids["interface_z_name"]
            )
            if not interface_z:
                LOGGER.info(
                    "Unable to create Cable %s in %s, unable to find the interface %s %s",
                    item.get_unique_id(),
                    diffsync.name,
                    ids["device_z_name"],
                    ids["interface_z_name"],
                )
                return item

        if interface_a.connected_endpoint_type:
            LOGGER.info(
                "Unable to create Cable in %s, port %s %s is already connected",
                diffsync.name,
                ids["device_a_name"],
                ids["interface_a_name"],
            )
            return item

        if interface_z.connected_endpoint_type:
            LOGGER.info(
                "Unable to create Cable in %s, port %s %s is already connected",
                diffsync.name,
                ids["device_z_name"],
                ids["interface_z_name"],
            )
            return item

        try:
            cable = diffsync.netbox.dcim.cables.create(
                termination_a_type="dcim.interface",
                termination_b_type="dcim.interface",
                termination_a_id=interface_a.remote_id,
                termination_b_id=interface_z.remote_id,
            )
        except pynetbox.core.query.RequestError as exc:
            LOGGER.warning("Unable to create Cable %s in %s (%s)", ids, diffsync.name, exc.error)
            return item

        interface_a.connected_endpoint_type = "dcim.interface"
        interface_z.connected_endpoint_type = "dcim.interface"

        LOGGER.info("Created Cable %s (%s) in NetBox", item.get_unique_id(), cable.id)
        item.remote_id = cable.id

        return item

    def delete(self):
        """Do not Delete the Cable in NetBox, just print a warning message.

        Returns:
            NetboxCable: DiffSync object
        """
        LOGGER.warning(
            "Cable %s is present in %s but not in the Network, please delete it manually if it shouldn't be in %s",
            self.get_unique_id(),
            self.diffsync.name,
            self.diffsync.name,
        )

        # try:
        #     cable = self.diffsync.netbox.dcim.cables.get(self.remote_id)
        #     cable.delete()
        # except pynetbox.core.query.RequestError as exc:
        #

        super().delete()
        return self
